import copy
import httpx
import inspect
import json
import os
import pytest

import snisp

from . import DATA_DIR


# 4600: MarketInsufficientCredits
# 4601: MarketTradeNoPurchaseError
# 4602: MarketTradeNotSoldError
# 4604: MarketTradeUnitLimitError


class TestMarkets:

    agent = snisp.agent.Agent(symbol='testing', faction='testing')

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_repr(self, respx_mock):
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        ship = self.agent.fleet('TEST_SHIP_SYMBOL')
        assert repr(ship.markets) == f'Markets({ship.agent!r}, ' \
                                     f'{ship.location})'

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_iter(self, respx_mock):
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoints.json'), encoding='utf8')
        )
        markets_side_effect = MarketsSideEffect(waypoints_data)
        market_route = respx_mock.get('/systems/TEST-SYSTEM/waypoints')
        market_route.side_effect = markets_side_effect

        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        ship = self.agent.fleet('TEST_SHIP_SYMBOL')

        with markets_side_effect:
            for market in ship.markets:
                assert any(i.symbol == 'MARKETPLACE' for i in market.traits)
            assert len(list(ship.markets)) == 1

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_call(self, respx_mock):
        waypoint_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoint.json'), encoding='utf8')
        )
        waypoint_data['data']['traits'][0]['symbol'] = 'MARKETPLACE'
        waypoint = snisp.waypoints.Waypoint(self.agent, waypoint_data['data'])
        market_side_effect = MarketSideEffect(waypoint_data)
        market_route = respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-WAYPOINT/market'
        )
        market_route.side_effect = market_side_effect
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        ship = self.agent.fleet('TEST_SHIP_SYMBOL')
        market = ship.markets()

        assert market == market.data
        assert market == market.data.data

        assert len(market) == len(market.to_dict())
        assert 'to_dict' in dir(market)
        assert market == ship.markets()

        assert ship.markets(waypoint_symbol='TEST-SYSTEM-WAYPOINT') == market
        assert ship.markets(waypoint=waypoint) == market
        with market_side_effect as tmp:
            tmp.invalid = True
            assert ship.markets().to_dict() == {}

        # 4603: MarketNotFoundError
        with market_side_effect as tmp:
            tmp.not_found = True
            with pytest.raises(snisp.exceptions.MarketNotFoundError):
                ship.markets()

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_search(self, respx_mock):
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoints.json'), encoding='utf8')
        )
        for wp in waypoints_data['data']:
            wp['traits'][0]['symbol'] = 'MARKETPLACE'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 1}
        ).mock(
            return_value=httpx.Response(200, json=waypoints_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 2}
        ).mock(
            return_value=httpx.Response(200, json={'data': []})
        )
        # lazy
        market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        iron_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        iron_market_data['data']['symbol'] = 'TEST-SYSTEM-WAYPOINT'
        for good in iron_market_data['data']['tradeGoods']:
            good['symbol'] = 'IRON'
        iron_market_data['data']['imports'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exchange'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exports'][0]['symbol'] = 'IRON'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-CLOSESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/'
            'TEST-SYSTEM-FARTHESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=iron_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-WAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=market_data)
        )
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        ship = self.agent.fleet('TEST_SHIP_SYMBOL')
        imports = ship.markets.search(imports='IRON')
        assert inspect.isgenerator(imports)
        imports = list(imports)
        assert len(imports) == 1
        assert imports[0][0].symbol == 'TEST-SYSTEM-FARTHESTWAYPOINT'
        assert snisp.markets.has_import(imports[0][1], 'IRON')

        exports = ship.markets.search(exports='PRECIOUS_STONES')
        assert inspect.isgenerator(exports)
        exports = list(exports)
        assert len(exports) == 2
        assert exports[0][0].symbol != 'TEST-SYSTEM-FARTHESTWAYPOINT'
        assert exports[1][0].symbol != 'TEST-SYSTEM-FARTHESTWAYPOINT'
        assert snisp.markets.has_export(exports[0][1], 'PRECIOUS_STONES')

        exchanges = ship.markets.search(exchanges='PRECIOUS_STONES')
        assert inspect.isgenerator(exchanges)
        exchanges = list(exchanges)
        assert len(exchanges) == 2
        assert exchanges[0][0].symbol != 'TEST-SYSTEM-FARTHESTWAYPOINT'
        assert exchanges[1][0].symbol != 'TEST-SYSTEM-FARTHESTWAYPOINT'
        assert snisp.markets.has_exchange(exchanges[0][1], 'PRECIOUS_STONES')

        invalid = ship.markets.search(imports='INVALID')
        assert inspect.isgenerator(invalid)
        assert len(list(invalid)) == 0

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_cheapest(self, respx_mock):
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoints.json'), encoding='utf8')
        )
        for wp in waypoints_data['data']:
            wp['traits'][0]['symbol'] = 'MARKETPLACE'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 1}
        ).mock(
            return_value=httpx.Response(200, json=waypoints_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 2}
        ).mock(
            return_value=httpx.Response(200, json={'data': []})
        )
        # lazy
        expensive_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        cheapest_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        for good in cheapest_market_data['data']['tradeGoods']:
            if good['type'] == 'IMPORT':
                good['purchasePrice'] = 1
        iron_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        iron_market_data['data']['symbol'] = 'TEST-SYSTEM-WAYPOINT'
        for good in iron_market_data['data']['tradeGoods']:
            good['symbol'] = 'IRON'
        iron_market_data['data']['imports'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exchange'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exports'][0]['symbol'] = 'IRON'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-CLOSESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=expensive_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/'
            'TEST-SYSTEM-FARTHESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=iron_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-WAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=cheapest_market_data)
        )
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        ship = self.agent.fleet('TEST_SHIP_SYMBOL')

        market = ship.markets.cheapest('PRECIOUS_STONES')
        assert market.symbol == 'TEST-SYSTEM-WAYPOINT'
        assert not ship.markets.cheapest('INVALID')

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_cheapest_import(self, respx_mock):
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoints.json'), encoding='utf8')
        )
        for wp in waypoints_data['data']:
            wp['traits'][0]['symbol'] = 'MARKETPLACE'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 1}
        ).mock(
            return_value=httpx.Response(200, json=waypoints_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 2}
        ).mock(
            return_value=httpx.Response(200, json={'data': []})
        )
        # lazy
        expensive_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        cheapest_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        for good in cheapest_market_data['data']['tradeGoods']:
            if good['type'] == 'IMPORT':
                good['sellPrice'] = 1
        iron_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        iron_market_data['data']['symbol'] = 'TEST-SYSTEM-WAYPOINT'
        for good in iron_market_data['data']['tradeGoods']:
            good['symbol'] = 'IRON'
        iron_market_data['data']['imports'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exchange'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exports'][0]['symbol'] = 'IRON'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-CLOSESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=expensive_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/'
            'TEST-SYSTEM-FARTHESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=iron_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-WAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=cheapest_market_data)
        )
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        ship = self.agent.fleet('TEST_SHIP_SYMBOL')

        market = ship.markets.cheapest_import('PRECIOUS_STONES')
        assert market[0].symbol == 'TEST-SYSTEM-WAYPOINT'
        assert not ship.markets.cheapest_import('INVALID')

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_cheapest_import_no_probed(self, respx_mock):
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoints.json'), encoding='utf8')
        )
        for wp in waypoints_data['data']:
            wp['traits'][0]['symbol'] = 'MARKETPLACE'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 1}
        ).mock(
            return_value=httpx.Response(200, json=waypoints_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 2}
        ).mock(
            return_value=httpx.Response(200, json={'data': []})
        )
        # lazy
        expensive_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        del expensive_market_data['data']['tradeGoods']
        cheapest_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        del cheapest_market_data['data']['tradeGoods']
        iron_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        iron_market_data['data']['symbol'] = 'TEST-SYSTEM-WAYPOINT'
        iron_market_data['data']['imports'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exchange'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exports'][0]['symbol'] = 'IRON'
        del iron_market_data['data']['tradeGoods']
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-CLOSESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=expensive_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/'
            'TEST-SYSTEM-FARTHESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=iron_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-WAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=cheapest_market_data)
        )
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        ship = self.agent.fleet('TEST_SHIP_SYMBOL')

        market = ship.markets.cheapest_import('PRECIOUS_STONES')
        assert market[0].symbol == 'TEST-SYSTEM-CLOSESTWAYPOINT'
        assert not ship.markets.cheapest_import('INVALID')

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_cheapest_export_or_exchange(self, respx_mock):
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoints.json'), encoding='utf8')
        )
        for wp in waypoints_data['data']:
            wp['traits'][0]['symbol'] = 'MARKETPLACE'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 1}
        ).mock(
            return_value=httpx.Response(200, json=waypoints_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 2}
        ).mock(
            return_value=httpx.Response(200, json={'data': []})
        )
        # lazy
        expensive_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        for good in expensive_market_data['data']['tradeGoods']:
            if good['type'] == 'EXPORT':
                good['symbol'] = 'IRON'
                good['purchasePrice'] = 1_000
                good['sellPrice'] = 1
        expensive_market_data['data']['exports'][0]['symbol'] = 'IRON'
        cheapest_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        for good in cheapest_market_data['data']['tradeGoods']:
            if good['type'] == 'EXPORT':
                good['purchasePrice'] = 1
        iron_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        iron_market_data['data']['symbol'] = 'TEST-SYSTEM-WAYPOINT'
        for good in iron_market_data['data']['tradeGoods']:
            good['symbol'] = 'IRON'
        iron_market_data['data']['imports'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exchange'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exports'][0]['symbol'] = 'IRON_ORE'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-CLOSESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=expensive_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/'
            'TEST-SYSTEM-FARTHESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=iron_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-WAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=cheapest_market_data)
        )
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        ship = self.agent.fleet('TEST_SHIP_SYMBOL')

        market = ship.markets.cheapest_export_or_exchange('IRON')
        assert market[0].symbol == 'TEST-SYSTEM-FARTHESTWAYPOINT'
        assert not ship.markets.cheapest_export_or_exchange('INVALID')

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_cheapest_export_or_exchange_no_probe(self, respx_mock):
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoints.json'), encoding='utf8')
        )
        for wp in waypoints_data['data']:
            wp['traits'][0]['symbol'] = 'MARKETPLACE'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 1}
        ).mock(
            return_value=httpx.Response(200, json=waypoints_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 2}
        ).mock(
            return_value=httpx.Response(200, json={'data': []})
        )
        # lazy
        expensive_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        del expensive_market_data['data']['tradeGoods']
        expensive_market_data['data']['exports'][0]['symbol'] = 'IRON'
        cheapest_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        del cheapest_market_data['data']['tradeGoods']
        iron_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        del iron_market_data['data']['tradeGoods']
        iron_market_data['data']['symbol'] = 'TEST-SYSTEM-WAYPOINT'
        iron_market_data['data']['imports'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exchange'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exports'][0]['symbol'] = 'IRON_ORE'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-CLOSESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=expensive_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/'
            'TEST-SYSTEM-FARTHESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=iron_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-WAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=cheapest_market_data)
        )
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        ship = self.agent.fleet('TEST_SHIP_SYMBOL')

        market = ship.markets.cheapest_export_or_exchange('IRON')
        assert market[0].symbol == 'TEST-SYSTEM-CLOSESTWAYPOINT'
        assert not ship.markets.cheapest_export_or_exchange('INVALID')

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_cheapest_exchange(self, respx_mock):
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoints.json'), encoding='utf8')
        )
        for wp in waypoints_data['data']:
            wp['traits'][0]['symbol'] = 'MARKETPLACE'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 1}
        ).mock(
            return_value=httpx.Response(200, json=waypoints_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 2}
        ).mock(
            return_value=httpx.Response(200, json={'data': []})
        )
        # lazy
        expensive_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        cheapest_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        for good in cheapest_market_data['data']['tradeGoods']:
            if good['type'] == 'EXPORT':
                good['purchasePrice'] = 1
        iron_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        iron_market_data['data']['symbol'] = 'TEST-SYSTEM-WAYPOINT'
        for good in iron_market_data['data']['tradeGoods']:
            good['symbol'] = 'IRON'
        iron_market_data['data']['imports'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exchange'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exports'][0]['symbol'] = 'IRON'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-CLOSESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=expensive_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/'
            'TEST-SYSTEM-FARTHESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=iron_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-WAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=cheapest_market_data)
        )
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        ship = self.agent.fleet('TEST_SHIP_SYMBOL')

        market = ship.markets.cheapest_exchange('PRECIOUS_STONES')
        assert market[0].symbol == 'TEST-SYSTEM-CLOSESTWAYPOINT'
        assert not ship.markets.cheapest_exchange('INVALID')

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_cheapest_exchange_no_probe(self, respx_mock):
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoints.json'), encoding='utf8')
        )
        for wp in waypoints_data['data']:
            wp['traits'][0]['symbol'] = 'MARKETPLACE'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 1}
        ).mock(
            return_value=httpx.Response(200, json=waypoints_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 2}
        ).mock(
            return_value=httpx.Response(200, json={'data': []})
        )
        # lazy
        expensive_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        del expensive_market_data['data']['tradeGoods']
        cheapest_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        del cheapest_market_data['data']['tradeGoods']
        iron_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        del iron_market_data['data']['tradeGoods']
        iron_market_data['data']['symbol'] = 'TEST-SYSTEM-WAYPOINT'
        iron_market_data['data']['imports'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exchange'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exports'][0]['symbol'] = 'IRON'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-CLOSESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=expensive_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/'
            'TEST-SYSTEM-FARTHESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=iron_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-WAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=cheapest_market_data)
        )
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        ship = self.agent.fleet('TEST_SHIP_SYMBOL')

        market = ship.markets.cheapest_exchange('PRECIOUS_STONES')
        assert market[0].symbol == 'TEST-SYSTEM-CLOSESTWAYPOINT'
        assert not ship.markets.cheapest_exchange('INVALID')

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_most_expensive_import(self, respx_mock):
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoints.json'), encoding='utf8')
        )
        for wp in waypoints_data['data']:
            wp['traits'][0]['symbol'] = 'MARKETPLACE'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 1}
        ).mock(
            return_value=httpx.Response(200, json=waypoints_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 2}
        ).mock(
            return_value=httpx.Response(200, json={'data': []})
        )
        # lazy
        expensive_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        del expensive_market_data['data']['tradeGoods']
        cheapest_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        del cheapest_market_data['data']['tradeGoods']
        iron_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        del iron_market_data['data']['tradeGoods']
        iron_market_data['data']['symbol'] = 'TEST-SYSTEM-WAYPOINT'
        iron_market_data['data']['imports'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exchange'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exports'][0]['symbol'] = 'IRON'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-CLOSESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=expensive_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/'
            'TEST-SYSTEM-FARTHESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=iron_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-WAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=cheapest_market_data)
        )
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        ship = self.agent.fleet('TEST_SHIP_SYMBOL')

        market = ship.markets.most_expensive_import('PRECIOUS_STONES')
        assert market[0].symbol == 'TEST-SYSTEM-CLOSESTWAYPOINT'
        assert not ship.markets.most_expensive_import('INVALID')

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_most_expensive_import_no_probe(self, respx_mock):
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoints.json'), encoding='utf8')
        )
        for wp in waypoints_data['data']:
            wp['traits'][0]['symbol'] = 'MARKETPLACE'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 1}
        ).mock(
            return_value=httpx.Response(200, json=waypoints_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 2}
        ).mock(
            return_value=httpx.Response(200, json={'data': []})
        )
        # lazy
        expensive_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        cheapest_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        for good in cheapest_market_data['data']['tradeGoods']:
            if good['type'] == 'EXPORT':
                good['purchasePrice'] = 1
        iron_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        iron_market_data['data']['symbol'] = 'TEST-SYSTEM-WAYPOINT'
        for good in iron_market_data['data']['tradeGoods']:
            good['symbol'] = 'IRON'
        iron_market_data['data']['imports'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exchange'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exports'][0]['symbol'] = 'IRON'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-CLOSESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=expensive_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/'
            'TEST-SYSTEM-FARTHESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=iron_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-WAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=cheapest_market_data)
        )
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        ship = self.agent.fleet('TEST_SHIP_SYMBOL')

        market = ship.markets.most_expensive_import('PRECIOUS_STONES')
        assert market[0].symbol == 'TEST-SYSTEM-CLOSESTWAYPOINT'
        assert not ship.markets.most_expensive_import('INVALID')

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_cheapest_export(self, respx_mock):
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoints.json'), encoding='utf8')
        )
        for wp in waypoints_data['data']:
            wp['traits'][0]['symbol'] = 'MARKETPLACE'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 1}
        ).mock(
            return_value=httpx.Response(200, json=waypoints_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 2}
        ).mock(
            return_value=httpx.Response(200, json={'data': []})
        )
        # lazy
        expensive_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        cheapest_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        for good in cheapest_market_data['data']['tradeGoods']:
            if good['type'] == 'EXPORT':
                good['purchasePrice'] = 1
        iron_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        iron_market_data['data']['symbol'] = 'TEST-SYSTEM-WAYPOINT'
        for good in iron_market_data['data']['tradeGoods']:
            good['symbol'] = 'IRON'
        iron_market_data['data']['imports'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exchange'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exports'][0]['symbol'] = 'IRON'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-CLOSESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=expensive_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/'
            'TEST-SYSTEM-FARTHESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=iron_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-WAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=cheapest_market_data)
        )
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        ship = self.agent.fleet('TEST_SHIP_SYMBOL')

        market = ship.markets.cheapest_export('PRECIOUS_STONES')
        assert market[0].symbol == 'TEST-SYSTEM-WAYPOINT'
        assert not ship.markets.cheapest_export('INVALID')

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_cheapest_export_no_probe(self, respx_mock):
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoints.json'), encoding='utf8')
        )
        for wp in waypoints_data['data']:
            wp['traits'][0]['symbol'] = 'MARKETPLACE'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 1}
        ).mock(
            return_value=httpx.Response(200, json=waypoints_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 2}
        ).mock(
            return_value=httpx.Response(200, json={'data': []})
        )
        # lazy
        expensive_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        del expensive_market_data['data']['tradeGoods']
        cheapest_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        del cheapest_market_data['data']['tradeGoods']
        iron_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        del iron_market_data['data']['tradeGoods']
        iron_market_data['data']['symbol'] = 'TEST-SYSTEM-WAYPOINT'
        iron_market_data['data']['imports'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exchange'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exports'][0]['symbol'] = 'IRON'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-CLOSESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=expensive_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/'
            'TEST-SYSTEM-FARTHESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=iron_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-WAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=cheapest_market_data)
        )
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        ship = self.agent.fleet('TEST_SHIP_SYMBOL')

        market = ship.markets.cheapest_export('PRECIOUS_STONES')
        assert market[0].symbol == 'TEST-SYSTEM-CLOSESTWAYPOINT'
        assert not ship.markets.cheapest_export('INVALID')

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_most_expensive_export(self, respx_mock):
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoints.json'), encoding='utf8')
        )
        for wp in waypoints_data['data']:
            wp['traits'][0]['symbol'] = 'MARKETPLACE'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 1}
        ).mock(
            return_value=httpx.Response(200, json=waypoints_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 2}
        ).mock(
            return_value=httpx.Response(200, json={'data': []})
        )
        # lazy
        expensive_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        cheapest_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        for good in cheapest_market_data['data']['tradeGoods']:
            if good['type'] == 'EXPORT':
                good['purchasePrice'] = 1
        iron_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        iron_market_data['data']['symbol'] = 'TEST-SYSTEM-WAYPOINT'
        for good in iron_market_data['data']['tradeGoods']:
            good['symbol'] = 'IRON'
        iron_market_data['data']['imports'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exchange'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exports'][0]['symbol'] = 'IRON'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-CLOSESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=expensive_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/'
            'TEST-SYSTEM-FARTHESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=iron_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-WAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=cheapest_market_data)
        )
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        ship = self.agent.fleet('TEST_SHIP_SYMBOL')

        market = ship.markets.most_expensive_export('PRECIOUS_STONES')
        assert market[0].symbol == 'TEST-SYSTEM-CLOSESTWAYPOINT'
        assert not ship.markets.most_expensive_export('INVALID')

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_most_expensive_export_no_probe(self, respx_mock):
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoints.json'), encoding='utf8')
        )
        for wp in waypoints_data['data']:
            wp['traits'][0]['symbol'] = 'MARKETPLACE'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 1}
        ).mock(
            return_value=httpx.Response(200, json=waypoints_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 2}
        ).mock(
            return_value=httpx.Response(200, json={'data': []})
        )
        # lazy
        expensive_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        del expensive_market_data['data']['tradeGoods']
        cheapest_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        del cheapest_market_data['data']['tradeGoods']
        iron_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        del iron_market_data['data']['tradeGoods']
        iron_market_data['data']['symbol'] = 'TEST-SYSTEM-WAYPOINT'
        iron_market_data['data']['imports'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exchange'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exports'][0]['symbol'] = 'IRON'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-CLOSESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=expensive_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/'
            'TEST-SYSTEM-FARTHESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=iron_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-WAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=cheapest_market_data)
        )
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        ship = self.agent.fleet('TEST_SHIP_SYMBOL')

        market = ship.markets.most_expensive_export('PRECIOUS_STONES')
        assert market[0].symbol == 'TEST-SYSTEM-CLOSESTWAYPOINT'
        assert not ship.markets.most_expensive_export('INVALID')

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_sells(self, respx_mock):
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoints.json'), encoding='utf8')
        )
        for wp in waypoints_data['data']:
            wp['traits'][0]['symbol'] = 'MARKETPLACE'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 1}
        ).mock(
            return_value=httpx.Response(200, json=waypoints_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 2}
        ).mock(
            return_value=httpx.Response(200, json={'data': []})
        )
        import_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        for good in import_market_data['data']['exports']:
            good['symbol'] = 'IRON'
        for good in import_market_data['data']['exchange']:
            good['symbol'] = 'IRON'
        import_market_data['data']['imports'][0]['symbol'] = 'PRECIOUS_STONES'
        export_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        for good in export_market_data['data']['imports']:
            good['symbol'] = 'IRON'
        for good in export_market_data['data']['exchange']:
            good['symbol'] = 'IRON'
        export_market_data['data']['exports'][0]['symbol'] = 'PRECIOUS_STONES'
        exchange_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        exchange_market_data['data']['symbol'] = 'TEST-SYSTEM-WAYPOINT'
        for good in exchange_market_data['data']['imports']:
            good['symbol'] = 'IRON'
        for good in exchange_market_data['data']['exports']:
            good['symbol'] = 'IRON'
        exchange_market_data['data']['exchange'][0]['symbol'] = 'PRECIOUS_STONES'  # noqa: E501
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-CLOSESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=import_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/'
            'TEST-SYSTEM-FARTHESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=exchange_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-WAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=export_market_data)
        )
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        ship = self.agent.fleet('TEST_SHIP_SYMBOL')

        markets = ship.markets.sells('PRECIOUS_STONES')
        assert inspect.isgenerator(markets)
        assert len(list(markets)) == 3
        assert not list(ship.markets.sells('INVALID'))

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_imports(self, respx_mock):
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoints.json'), encoding='utf8')
        )
        for wp in waypoints_data['data']:
            wp['traits'][0]['symbol'] = 'MARKETPLACE'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 1}
        ).mock(
            return_value=httpx.Response(200, json=waypoints_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 2}
        ).mock(
            return_value=httpx.Response(200, json={'data': []})
        )
        # lazy
        expensive_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        cheapest_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        for good in cheapest_market_data['data']['tradeGoods']:
            if good['type'] == 'EXPORT':
                good['purchasePrice'] = 1
        iron_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        iron_market_data['data']['symbol'] = 'TEST-SYSTEM-WAYPOINT'
        for good in iron_market_data['data']['tradeGoods']:
            good['symbol'] = 'IRON'
        iron_market_data['data']['imports'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exchange'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exports'][0]['symbol'] = 'IRON'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-CLOSESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=expensive_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/'
            'TEST-SYSTEM-FARTHESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=iron_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-WAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=cheapest_market_data)
        )
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        ship = self.agent.fleet('TEST_SHIP_SYMBOL')

        markets = ship.markets.imports('PRECIOUS_STONES')
        assert inspect.isgenerator(markets)
        assert len(list(markets)) == 2
        assert not list(ship.markets.imports('INVALID'))

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_exports(self, respx_mock):
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoints.json'), encoding='utf8')
        )
        for wp in waypoints_data['data']:
            wp['traits'][0]['symbol'] = 'MARKETPLACE'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 1}
        ).mock(
            return_value=httpx.Response(200, json=waypoints_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 2}
        ).mock(
            return_value=httpx.Response(200, json={'data': []})
        )
        # lazy
        expensive_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        cheapest_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        for good in cheapest_market_data['data']['tradeGoods']:
            if good['type'] == 'EXPORT':
                good['purchasePrice'] = 1
        iron_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        iron_market_data['data']['symbol'] = 'TEST-SYSTEM-WAYPOINT'
        for good in iron_market_data['data']['tradeGoods']:
            good['symbol'] = 'IRON'
        iron_market_data['data']['imports'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exchange'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exports'][0]['symbol'] = 'IRON'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-CLOSESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=expensive_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/'
            'TEST-SYSTEM-FARTHESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=iron_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-WAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=cheapest_market_data)
        )
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        ship = self.agent.fleet('TEST_SHIP_SYMBOL')

        markets = ship.markets.exports('PRECIOUS_STONES')
        assert inspect.isgenerator(markets)
        assert len(list(markets)) == 2
        assert not list(ship.markets.exports('INVALID'))

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_exchanges(self, respx_mock):
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoints.json'), encoding='utf8')
        )
        for wp in waypoints_data['data']:
            wp['traits'][0]['symbol'] = 'MARKETPLACE'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 1}
        ).mock(
            return_value=httpx.Response(200, json=waypoints_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 2}
        ).mock(
            return_value=httpx.Response(200, json={'data': []})
        )
        # lazy
        expensive_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        cheapest_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        for good in cheapest_market_data['data']['tradeGoods']:
            if good['type'] == 'EXCHANGE':
                good['purchasePrice'] = 1
        iron_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        iron_market_data['data']['symbol'] = 'TEST-SYSTEM-WAYPOINT'
        for good in iron_market_data['data']['tradeGoods']:
            good['symbol'] = 'IRON'
        iron_market_data['data']['imports'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exchange'][0]['symbol'] = 'IRON'
        iron_market_data['data']['exports'][0]['symbol'] = 'IRON'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-CLOSESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=expensive_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/'
            'TEST-SYSTEM-FARTHESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=iron_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-WAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=cheapest_market_data)
        )
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        ship = self.agent.fleet('TEST_SHIP_SYMBOL')

        markets = ship.markets.exchanges('PRECIOUS_STONES')
        assert inspect.isgenerator(markets)
        assert len(list(markets)) == 2
        assert not list(ship.markets.exchanges('INVALID'))

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_shipyard_market_data(self, respx_mock):
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        shipyards_data = json.load(
            open(os.path.join(DATA_DIR, 'shipyards.json'), encoding='utf8')
        )
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoints.json'), encoding='utf8')
        )
        shipyards_side_effect = ShipyardsSideEffect(
            data=shipyards_data, waypoints=waypoints_data
        )
        shipyards_route = respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-WAYPOINT/shipyard'
        )
        shipyards_route.side_effect = shipyards_side_effect
        waypoints_route = respx_mock.get('/systems/TEST-SYSTEM/waypoints')
        waypoints_route.side_effect = shipyards_side_effect.waypoints_side_effect  # noqa: E501

        with shipyards_side_effect as tmp:
            tmp.waypoints['data'][1]['traits'][0]['symbol'] = 'SHIPYARD'
            ship = self.agent.fleet('TEST_SHIP_SYMBOL')
            shipyard_data = ship.markets.shipyard_market_data('SHIP_PROBE')
            assert len(shipyard_data) == 2
            assert shipyard_data[0].symbol == 'TEST-SYSTEM-WAYPOINT'
            assert shipyard_data[1].type == 'SHIP_PROBE'

        with shipyards_side_effect as tmp:
            tmp.waypoints['data'][1]['traits'][0]['symbol'] = 'SHIPYARD'
            ship = self.agent.fleet('TEST_SHIP_SYMBOL')
            shipyard_data = ship.markets.shipyard_market_data('SHIP_ORE_HOUND')
            assert len(shipyard_data) == 2
            assert shipyard_data[0] is None
            assert shipyard_data[1] is None

        with shipyards_side_effect as tmp:
            tmp.waypoints['data'][1]['traits'][0]['symbol'] = 'SHIPYARD'
            ship = self.agent.fleet('TEST_SHIP_SYMBOL')
            with pytest.raises(snisp.exceptions.SpaceAttributeError):
                shipyard_data = ship.markets.shipyard_market_data('INVALID')

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_best_market_pairs(self, respx_mock):
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoints.json'), encoding='utf8')
        )
        for wp in waypoints_data['data']:
            wp['traits'][0]['symbol'] = 'MARKETPLACE'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 1}
        ).mock(
            return_value=httpx.Response(200, json=waypoints_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints', params={'page': 2}
        ).mock(
            return_value=httpx.Response(200, json={'data': []})
        )
        # lazy
        expensive_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        cheapest_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        for good in cheapest_market_data['data']['tradeGoods']:
            if good['symbol'] == 'PRECIOUS_STONES':
                good['purchasePrice'] = 100
        for good in expensive_market_data['data']['tradeGoods']:
            if good['symbol'] == 'PRECIOUS_STONES':
                good['purchasePrice'] = 1_000
                good['sellPrice'] = 1_000
        cheapest_market_data['data']['symbol'] = 'TEST-SYSTEM-WAYPOINT'
        expensive_market_data['data']['symbol'] = 'TEST-SYSTEM-CLOSESTWAYPOINT'
        best_sell_market_data = json.load(
            open(os.path.join(DATA_DIR, 'market_data.json'), encoding='utf8')
        )
        best_sell_market_data['data']['symbol'] = 'TEST-SYSTEM-FARTHESTWAYPOINT'  # noqa: E501
        for good in best_sell_market_data['data']['tradeGoods']:
            if good['symbol'] == 'PRECIOUS_STONES':
                good['sellPrice'] = 2_000
        best_sell_market_data['data']['exchange'][0]['symbol'] = 'IRON'
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-WAYPOINT'
        ).mock(
            return_value=httpx.Response(
                200, json={'data': waypoints_data['data'][1]}
            )
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-FARTHESTWAYPOINT'
        ).mock(
            return_value=httpx.Response(
                200, json={'data': waypoints_data['data'][2]}
            )
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-CLOSESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=expensive_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/'
            'TEST-SYSTEM-FARTHESTWAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=best_sell_market_data)
        )
        respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-WAYPOINT/market'
        ).mock(
            return_value=httpx.Response(200, json=cheapest_market_data)
        )
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        ship = self.agent.fleet('TEST_SHIP_SYMBOL')

        market_data = [m.data for m in ship.markets]
        pairs = snisp.markets.best_market_pairs(ship, market_data)
        assert int(pairs[0].distance) == 70
        assert pairs[0].trade_symbol == 'PRECIOUS_STONES'
        assert pairs[0].import_market.symbol == 'TEST-SYSTEM-FARTHESTWAYPOINT'
        assert pairs[0].export_market.symbol == 'TEST-SYSTEM-WAYPOINT'
        assert not snisp.markets.best_market_pairs(
            ship, market_data, price_delta=1_000_000
        )


class MarketSideEffect:

    def __init__(self, data):
        self.orig_data = data
        self.data = copy.deepcopy(self.orig_data)
        self.invalid = False
        self.not_found = False

    def __call__(self, request, route):
        if self.invalid:
            return httpx.Response(
                400, json={'error': {'data': {'code': 404}}}
            )
        if self.not_found:
            return httpx.Response(
                400, json={'error': {'data': {'code': 4603}}}
            )
        return httpx.Response(200, json=self.data)

    def __enter__(self):
        self.reset()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.reset()

    def reset(self):
        self.data = copy.deepcopy(self.orig_data)
        self.invalid = False
        self.not_found = False


class MarketsSideEffect:

    def __init__(self, data):
        self.orig_data = data
        self.data = copy.deepcopy(self.orig_data)

    def __call__(self, request, route):
        if int(request.url.params.get('page', 1)) > 1:
            return httpx.Response(200, json={'data': []})
        traits = request.url.params.get('traits')
        output = []
        for wp in self.data['data']:
            if any(i['symbol'] == traits for i in wp['traits']):
                output.append(wp)
        self.data['data'] = output
        return httpx.Response(200, json=self.data)

    def __enter__(self):
        self.reset()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.reset()

    def reset(self):
        self.data = copy.deepcopy(self.orig_data)


class ShipyardsSideEffect:

    def __init__(self, *, data, waypoints):
        self.orig_data = data
        self.data = copy.deepcopy(self.orig_data)
        self.orig_waypoints = waypoints
        self.waypoints = copy.deepcopy(self.orig_waypoints)

    def __call__(self, request, route):
        return httpx.Response(200, json=self.data)

    def waypoints_side_effect(self, request, route):
        if int(request.url.params.get('page', 1)) > 1:
            return httpx.Response(200, json={'data': []})
        traits = request.url.params.get('traits')
        output = []
        for wp in self.waypoints['data']:
            if any(i['symbol'] == traits for i in wp['traits']):
                output.append(wp)
        self.waypoints['data'] = output
        return httpx.Response(200, json=self.waypoints)

    def __enter__(self):
        self.reset()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.reset()

    def reset(self):
        self.data = copy.deepcopy(self.orig_data)
        self.waypoints = copy.deepcopy(self.orig_waypoints)



import dateutil
import functools
import httpx
import itertools
import logging
import threading
import time

import snisp

from datetime import datetime, timezone


logger = logging.getLogger(__name__)


class CachedRateLimiter:

    def __init__(self):
        self.lock = threading.Lock()
        self.last_request = datetime.now(timezone.utc)

    def __call__(self, func):

        @functools.wraps(func)
        def inner(*args, **kwargs):
            if args[0].testing:
                return func(*args, **kwargs)
            is_get = True if func.__name__ == 'get' else False
            with self.lock:
                if is_get:
                    if cache := snisp.cache.lookup(*args, **kwargs):
                        return cache
                # ignoring bursts
                delta = datetime.now(timezone.utc) - self.last_request
                if not delta.seconds:
                    second = pow(10, -6) * delta.microseconds
                    if second < .5:
                        time.sleep(.5 - second)
                try:
                    response = func(*args, **kwargs)
                    if is_get:
                        snisp.cache.insert(response, *args, **kwargs)
                    return response
                finally:
                    self.last_request = datetime.now(timezone.utc)
        return inner


def cooldown(func):

    @functools.wraps(func)
    def inner(*args, **kwargs):
        for ship in itertools.chain(args, kwargs.values()):
            if isinstance(ship, snisp.fleet.Ship):
                if expires := ship.cooldown.expiration:
                    expires = dateutil.parser.parse(expires)
                    delta = expires - datetime.now(timezone.utc)
                    if delta.days >= 0:  # pragma: no cover aka impatient
                        time.sleep(delta.seconds)
        return func(*args, **kwargs)
    return inner


def docked(func):

    @functools.wraps(func)
    def inner(*args, **kwargs):
        for ship in itertools.chain(args, kwargs.values()):
            if isinstance(ship, snisp.fleet.Ship):
                if ship.nav.status != 'DOCKED':
                    ship.dock()
        return func(*args, **kwargs)
    return inner


def in_orbit(func):

    @functools.wraps(func)
    def inner(*args, **kwargs):
        for ship in itertools.chain(args, kwargs.values()):
            if isinstance(ship, snisp.fleet.Ship):
                if ship.nav.status != 'IN_ORBIT':
                    ship.orbit()
        return func(*args, **kwargs)
    return inner


def retry(jitter=.2, max_retries=5):

    def wrapper(func):

        @functools.wraps(func)
        def inner(*args, **kwargs):
            ship = args[0]
            for attempt in range(1, max_retries + 1):
                try:
                    return func(*args, **kwargs)
                except httpx.HTTPError as e:
                    logger.warning(
                        f'Attempt: {attempt}/{max_retries}. '
                        f'Received {e!r}.'
                    )
                    if not ship.agent.client.testing:  # pragma: no cover
                        time.sleep(jitter * attempt)
                    last_exception = e
                except snisp.exceptions.ClientError as e:
                    logger.warning(
                        f'Attempt: {attempt}/{max_retries}. '
                        f'Received: {e!r}.'
                    )
                    if data := e.data:
                        if cooldown := data.get(
                            'cooldown'
                        ):  # pragma: no cover
                            if wait := cooldown.get('remainingSeconds'):
                                if not ship.agent.client.testing:
                                    time.sleep(float(wait))
                        elif wait := max(
                            [
                                data.get('secondsToArrival', 0),
                                data.get('retryAfter', 0),
                                data.get('remainingSeconds', 0),
                            ]
                        ):  # pragma: no cover
                            if not ship.agent.client.testing:
                                time.sleep(float(wait))
                        elif code := data.get('code'):
                            if err := snisp.exceptions.error_codes.get(
                                int(code)
                            ):
                                raise err(str(e))
                    last_exception = e
            raise last_exception
        return inner
    return wrapper


def transit(func):

    @functools.wraps(func)
    def inner(*args, **kwargs):
        for ship in itertools.chain(args, kwargs.values()):
            if isinstance(ship, snisp.fleet.Ship):
                if ship.nav.status == 'IN_TRANSIT':
                    if arrives_in := ship.arrival:  # pragma: no cover
                        time.sleep(arrives_in)
                    ship.arrived_at_destination()
        return func(*args, **kwargs)
    return inner



import functools
import logging

from collections import Counter

from snisp import exceptions, utils, systems
from snisp.decorators import cooldown, docked, in_orbit, retry, transit
from snisp.exceptions import ClientError
from snisp.shipyards import Shipyard


logger = logging.getLogger(__name__)


class Waypoints:

    def __init__(self, agent, location):
        self.agent = agent
        self.location = location

    def __repr__(self):
        cls = self.__class__.__name__
        return f'{cls}({self.agent!r}, {self.location!r})'

    def __iter__(self):
        """
        Iterates over all of the Waypoint's in the Ship's current location

        Yields:
            Waypoint or a subclass of Waypoint
        """
        page = 1
        response = self.get_page(page=page)
        while data := response.json()['data']:
            for waypoint in data:
                cls = class_factory(waypoint['type'])
                yield cls(self.agent, waypoint)
            page += 1
            response = self.get_page(page=page)

    def __call__(
        self,
        *,
        system_symbol=None,
        traits=None,
        types=None,
        filters=None,
        page=1,
    ):
        """
        Iterates over Waypoints while yield only the ones that match the
        filters

        Kwargs:
            system_symbol: Symbol of the System to search. If not specified,
                           the System in the current self.location will be used
                           Default is None.
            traits: Single trait to filter on. See snisp.utils.WAYPOINT_TRAITS
                    for all traits
                    Default is None
            types: Single type to filter on. See snisp.utils.WAYPOINT_TYPES
                   for all types
                   Default is None
            filters: A dictionary of top-level items to filter the Waypoints.
                     The key, value pairs in the Waypoint's waypoint.to_dict()
                     data is compared against the filters
                     e.g., a filter of {'orbitals': []} would return only
                     Waypoints that don't have any orbitals
                     Default is None

        Yields:
            Waypoint or a subclass of Waypoint
        """
        filters = dict(filters) if filters is not None else {}
        filters = {utils.camel_case(k): v for k, v in filters.items()}
        response = self.get_page(
            system_symbol=system_symbol,
            traits=traits,
            types=types,
            page=page,
        )
        while data := response.json()['data']:
            for waypoint in data:
                if all(
                    False for k, v in filters.items()
                    if waypoint.get(k) != v
                ):
                    cls = class_factory(waypoint['type'])
                    yield cls(self.agent, waypoint)
            page += 1
            response = self.get_page(
                system_symbol=system_symbol,
                traits=traits,
                types=types,
                page=page,
            )

    @retry()
    def get_page(self, *, page=1, system_symbol=None, traits=None, types=None):
        params = {'page': int(page)}
        if types is not None:
            types = types.strip().upper()
            if types not in utils.WAYPOINT_TYPES:
                raise exceptions.SpaceAttributeError(
                    f'{types} is not an acceptable Waypoint type. '
                    'See snisp.utils.WAYPOINT_TYPES for acceptable types.'
                )
            params['type'] = types
        if traits is not None:
            traits = traits.strip().upper()
            if traits not in utils.WAYPOINT_TRAITS:
                raise exceptions.SpaceAttributeError(
                    f'{traits} is not an acceptable Waypoint trait. '
                    'See snisp.utils.WAYPOINT_TRAITS for acceptable traits.'
                )
            params['traits'] = traits
        if system_symbol is None:
            system_symbol = self.location.system
        return self.agent.client.get(
            f'/systems/{system_symbol}/waypoints', params=params
        )

    @retry()
    def get(self, *, waypoint=None, system_symbol=None, waypoint_symbol=None):
        """
        Gets a Waypoint or Waypoint subclass.

        Convenient for converting a something back to a Waypoint
        or Waypoint subclass

        >>> ship = next(iter(agent.fleet))
        >>> market = next(iter(ship.markets))
        >>> type(market)
        <class 'snisp.markets.Market'>
        >>> waypoint = ship.waypoints.get(waypoint=market)
        >>> type(waypoint)
        <class 'snisp.waypoints.EngineeredAsteroid'>

        Kwargs:
            waypoint: Waypoint or Waypoint subclass. Default is None
            system_symbol: Symbol of a System to use. Default is None
            waypoint_symbol: Symbol of a Waypoint to use. Default is None

        Returns:
            Waypoint or Waypoint subclass
        """
        if waypoint is not None:
            system_symbol = waypoint.system_symbol
            waypoint_symbol = waypoint.symbol
        else:
            if system_symbol is None:
                system_symbol = self.location.system
            if waypoint_symbol is None:
                waypoint_symbol = self.location.waypoint
        response = self.agent.client.get(
            f'/systems/{system_symbol}/waypoints/{waypoint_symbol}'
        )
        data = response.json()['data']
        cls = class_factory(data['type'])
        return cls(self.agent, data)

    @retry()
    def refresh(self, waypoint):
        """Returns a new Waypoint class or subclass object of the
        current Waypoint

        Returns:
            A new Waypoint or Waypoint subclass instance from self
        """
        response = self.agent.client.get(
            f'/systems/{waypoint.system_symbol}/waypoints/{waypoint.symbol}'
        )
        data = response.json()['data']
        data['systemSymbol'] = waypoint.system_symbol
        data['symbol'] = waypoint.symbol
        cls = class_factory(data['type'])
        return cls(self.agent, data)

    @retry()
    def search(
        self, *, traits=None, types=None, system_symbol=None, filters=None,
    ):
        """Convenience method. Equivalent to __call__

        Kwargs:
            system_symbol: Symbol of the System to search. If not specified,
                           the System in the current self.location will be used
                           Default is None.
            traits: Single trait to filter on. See snisp.utils.WAYPOINT_TRAITS
                    for all traits
                    Default is None
            types: Single type to filter on. See snisp.utils.WAYPOINT_TYPES
                   for all types
                   Default is None
            filters: A dictionary of top-level items to filter the Waypoints.
                     The key, value pairs in the Waypoint's waypoint.to_dict()
                     data is compared against the filters
                     e.g., a filter of {'orbitals': []} would return only
                     Waypoints that don't have any orbitals
                     Default is None

        Yields:
            Waypoint or Waypoint subclass
        """
        yield from self(
            system_symbol=system_symbol,
            traits=traits,
            types=types,
            filters=filters,
        )

    @retry()
    @transit
    def chart(self, ship):
        """
        Charts the Waypoint at the ship's current location

        Args:
            ship: Ship to use. All snisp.fleet.Ship objects will have this
                  automatically monkey-patched in.
        Blocks:
            True: Won't be executed until Ship reaches destination

        Returns:
            Chart
        """
        try:
            response = self.agent.client.post(
                f'/my/ships/{ship.symbol}/chart'
            )
        except ClientError as e:
            if data := e.data:
                if data.get('code') == 4230:
                    logger.warning(
                        f'Waypoint at {ship.symbol} has already been charted'
                    )
                    return Chart(self.agent, {})
            raise e
        logger.info(
            f'{ship.registration.role}: {ship.symbol} | '
            f'Charted {ship.nav.waypoint_symbol}'
        )
        return Chart(self.agent, response.json()['data'])

    @retry()
    @transit
    @cooldown
    @in_orbit
    def scan(self, ship):
        """
        Scans the Waypoints in the ship's current System

        Args:
            ship: Ship to use. All snisp.fleet.Ship objects will have this
                  automatically monkey-patched in.
        Blocks:
            True: Won't be executed until Ship reaches destination and/or
                  the cooldown period has passed

        Yields:
            Waypoint or Waypoint subclasses
        """
        response = self.agent.client.post(
            f'/my/ships/{ship.symbol}/scan/waypoints'
        )
        data = response.json()['data']
        ship.update_data_item('cooldown', data['cooldown'])
        for waypoint in data['waypoints']:
            cls = class_factory(waypoint['type'])
            yield cls(self.agent, waypoint)

    def surveys(self):
        surveyed_wps = set()
        for ship in self.agent.fleet:
            if ship.nav.waypoint_symbol in surveyed_wps:
                continue
            if ship.nav.route.destination.type in utils.SURVEYABLE_WAYPOINTS:
                if ship.can_survey and ship.nav.status == 'IN_ORBIT':
                    try:
                        yield ship.waypoints.survey()
                    except Exception as e:
                        logger.warning(e)
                    else:
                        surveyed_wps.add(ship.nav.waypoint_symbol)

    @retry()
    @transit
    @cooldown
    @in_orbit
    def survey(self, ship):
        """
        Surveys the Waypoint at the ship's current location

        Args:
            ship: Ship to use. All snisp.fleet.Ship objects will have this
                  automatically monkey-patched in.
        Blocks:
            True: Won't be executed until Ship reaches destination and/or
                  the cooldown period has passed

        Returns:
            Survey
        """
        if ship.nav.route.destination.type not in utils.SURVEYABLE_WAYPOINTS:
            raise exceptions.ShipSurveyWaypointTypeError(
                f'{ship.nav.route.destination.type} is not an '
                'acceptable Waypoint type for surveying. '
                'See snisp.utils.SURVEYABLE_WAYPOINTS for acceptable types.'
            )
        response = self.agent.client.post(f'/my/ships/{ship.symbol}/survey')
        data = response.json()['data']
        ship.update_data_item('cooldown', data['cooldown'])
        logger.info(
            f'{ship.registration.role}: {ship.symbol} | '
            f'Surveyed {ship.nav.waypoint_symbol}'
        )
        return Survey(self.agent, data)

    def construction_sites(self, *, system_symbol=None):
        """
        Convenience method for finding ConstructionSites

        Kwargs:
            system_symbol: System to search. Default is None for current System

        Yields:
            ConstructionSite
        """
        for waypoint in self(
            system_symbol=system_symbol
        ):
            if waypoint.is_under_construction:
                yield ConstructionSite.get(self.agent, waypoint)

    def shipyards(self, *, system_symbol=None):
        """
        Convenience method for finding Shipyards

        Kwargs:
            system_symbol: System to search. Default is None for current System

        Yields:
            Shipyard
        """
        for shipyard in self(system_symbol=system_symbol, traits='SHIPYARD'):
            yield Shipyard(self.agent, shipyard.to_dict())

    def planets(self, *, system_symbol=None, traits=None):
        """
        Convenience method for finding Planets

        Kwargs:
            system_symbol: System to search. Default is None for current System
            traits: Single trait to filter on. See snisp.utils.WAYPOINT_TRAITS
                    for all traits
                    Default is None

        Yields:
            Planet
        """
        for planet in self(
            system_symbol=system_symbol, types='PLANET', traits=traits
        ):
            yield Planet(self.agent, planet.to_dict())

    def gas_giants(self, *, system_symbol=None, traits=None):
        """
        Convenience method for finding Gas Giants

        Kwargs:
            system_symbol: System to search. Default is None for current System
            traits: Single trait to filter on. See snisp.utils.WAYPOINT_TRAITS
                    for all traits
                    Default is None

        Yields:
            GasGiant
        """
        for gas_giant in self(
            system_symbol=system_symbol, types='GAS_GIANT', traits=traits
        ):
            yield GasGiant(self.agent, gas_giant.to_dict())

    def moons(self, *, system_symbol=None, traits=None):
        """
        Convenience method for finding Moons

        Kwargs:
            system_symbol: System to search. Default is None for current System
            traits: Single trait to filter on. See snisp.utils.WAYPOINT_TRAITS
                    for all traits
                    Default is None

        Yields:
            Moon
        """
        for moon in self(
            system_symbol=system_symbol, types='MOON', traits=traits
        ):
            yield Moon(self.agent, moon.to_dict())

    def orbital_stations(self, *, system_symbol=None, traits=None):
        """
        Convenience method for finding Orbital Stations

        Kwargs:
            system_symbol: System to search. Default is None for current System
            traits: Single trait to filter on. See snisp.utils.WAYPOINT_TRAITS
                    for all traits
                    Default is None

        Yields:
            OrbitalStation
        """
        for orbital_station in self(
            system_symbol=system_symbol, types='ORBITAL_STATION', traits=traits
        ):
            yield OrbitalStation(self.agent, orbital_station.to_dict())

    def jump_gates(self, *, system_symbol=None, traits=None):
        """
        Convenience method for finding Jump Gates

        Kwargs:
            system_symbol: System to search. Default is None for current System
            traits: Single trait to filter on. See snisp.utils.WAYPOINT_TRAITS
                    for all traits
                    Default is None

        Yields:
            JumpGate
        """
        for jump_gate in self(
            system_symbol=system_symbol, types='JUMP_GATE', traits=traits
        ):
            yield JumpGate(self.agent, jump_gate.to_dict())

    def asteroid_fields(self, *, system_symbol=None, traits=None):
        """
        Convenience method for finding Asteroid Fields

        Kwargs:
            system_symbol: System to search. Default is None for current System
            traits: Single trait to filter on. See snisp.utils.WAYPOINT_TRAITS
                    for all traits
                    Default is None

        Yields:
            AsteroidField
        """
        for asteroid_field in self(
            system_symbol=system_symbol, types='ASTEROID_FIELD', traits=traits
        ):
            yield AsteroidField(self.agent, asteroid_field.to_dict())

    def asteroids(self, *, system_symbol=None, traits=None):
        """
        Convenience method for finding Asteroids

        Kwargs:
            system_symbol: System to search. Default is None for current System
            traits: Single trait to filter on. See snisp.utils.WAYPOINT_TRAITS
                    for all traits
                    Default is None

        Yields:
            Asteroid
        """
        for asteroid in self(
            system_symbol=system_symbol, types='ASTEROID', traits=traits
        ):
            yield Asteroid(self.agent, asteroid.to_dict())

    def engineered_asteroids(self, *, system_symbol=None, traits=None):
        """
        Convenience method for finding Engineered Asteroids

        Kwargs:
            system_symbol: System to search. Default is None for current System
            traits: Single trait to filter on. See snisp.utils.WAYPOINT_TRAITS
                    for all traits
                    Default is None

        Yields:
            EngineeredAsteroid
        """
        for engineered_asteroid in self(
            system_symbol=system_symbol,
            types='ENGINEERED_ASTEROID',
            traits=traits
        ):
            yield EngineeredAsteroid(self.agent, engineered_asteroid.to_dict())

    def asteroid_bases(self, *, system_symbol=None, traits=None):
        """
        Convenience method for finding Asteroid Bases

        Kwargs:
            system_symbol: System to search. Default is None for current System
            traits: Single trait to filter on. See snisp.utils.WAYPOINT_TRAITS
                    for all traits
                    Default is None

        Yields:
            AsteroidBase
        """
        for asteroid_base in self(
            system_symbol=system_symbol, types='ASTEROID_BASE', traits=traits
        ):
            yield AsteroidBase(self.agent, asteroid_base.to_dict())

    def nebulas(self, *, system_symbol=None, traits=None):
        """
        Convenience method for finding Nebulas

        Kwargs:
            system_symbol: System to search. Default is None for current System
            traits: Single trait to filter on. See snisp.utils.WAYPOINT_TRAITS
                    for all traits
                    Default is None

        Yields:
            Nebula
        """
        for nebula in self(
            system_symbol=system_symbol, types='NEBULA', traits=traits
        ):
            yield Nebula(self.agent, nebula.to_dict())

    def debris_fields(self, *, system_symbol=None, traits=None):
        """
        Convenience method for finding Debris Fields

        Kwargs:
            system_symbol: System to search. Default is None for current System
            traits: Single trait to filter on. See snisp.utils.WAYPOINT_TRAITS
                    for all traits
                    Default is None

        Yields:
            DebrisField
        """
        for debris_field in self(
            system_symbol=system_symbol, types='DEBRIS_FIELD', traits=traits
        ):
            yield DebrisField(self.agent, debris_field.to_dict())

    def gravity_wells(self, *, system_symbol=None, traits=None):
        """
        Convenience method for finding Gravity Wells

        Kwargs:
            system_symbol: System to search. Default is None for current System
            traits: Single trait to filter on. See snisp.utils.WAYPOINT_TRAITS
                    for all traits
                    Default is None

        Yields:
            GravityWell
        """
        for gravity_well in self(
            system_symbol=system_symbol, types='GRAVITY_WELL', traits=traits
        ):
            yield GravityWell(self.agent, gravity_well.to_dict())

    def artificial_gravity_wells(self, *, system_symbol=None, traits=None):
        """
        Convenience method for finding Artificial Gravity Wells

        Kwargs:
            system_symbol: System to search. Default is None for current System
            traits: Single trait to filter on. See snisp.utils.WAYPOINT_TRAITS
                    for all traits
                    Default is None

        Yields:
            ArtificialGravityWell
        """
        for artificial_gravity_well in self(
            system_symbol=system_symbol,
            types='ARTIFICIAL_GRAVITY_WELL',
            traits=traits
        ):
            yield ArtificialGravityWell(
                self.agent, artificial_gravity_well.to_dict()
            )


class Waypoint(utils.AbstractJSONItem):

    def __init__(self, agent, waypoint):
        self.agent = agent
        self._data = waypoint

    @property
    def data(self):
        response = self.agent.client.get(
            f'/systems/{self.system_symbol}/waypoints/{self.waypoint_symbol}'
        )
        return WaypointData(self.agent, response.json()['data'])

    @property
    def location(self):
        """Property for the Waypoints Location

        Returns:
            snisp.system.Location: Waypoints Location
        """
        return systems.Location(self.agent, self.to_dict())


class WaypointData(utils.AbstractJSONItem):

    def __init__(self, agent, data):
        self.agent = agent
        self._data = data


class Chart(utils.AbstractJSONItem):

    def __init__(self, agent, waypoint):
        self.agent = agent
        self._data = waypoint


class Survey(utils.AbstractJSONItem):

    def __init__(self, agent, waypoint):
        self.agent = agent
        self._data = waypoint

    def best(self, *trade_symbols):
        if not trade_symbols:
            trade_symbols = {
                d.symbol for s in self.surveys for d in s.deposits
            }
        else:
            trade_symbols = {i for i in trade_symbols}
        # TODO: If this is used elsewhere, make it an enum
        survey_sizes = {
            'SMALL': 0,
            'MODERATE': 1,
            'LARGE': 2,
        }
        counter = Counter()
        sig_to_survey = {i.signature: i for i in self.surveys}
        for survey in self.surveys:
            for deposit in survey.deposits:
                if deposit.symbol in trade_symbols:
                    counter[survey.signature] += 1
        best_survey = None
        prev_count = None
        for survey_sig, count in counter.most_common():
            survey = sig_to_survey[survey_sig]
            if prev_count is not None and count < prev_count:
                break
            if best_survey is None:
                best_survey = survey
            elif survey_sizes[survey.size] > survey_sizes[best_survey.size]:
                best_survey = survey
            prev_count = count
        return best_survey


class ConstructionSite(utils.AbstractJSONItem):

    def __init__(self, agent, waypoint):
        self.agent = agent
        self._data = waypoint

    @classmethod
    @retry()
    def get(cls, agent, waypoint):
        """
        Class method that creates a ConstructionSite instance from a Waypiont
        Useful for when going from JumpGate Waypoint type to
        a ConstrutionSite type, for instance

        Args:
            agent: Your Agent
            waypoint: The Waypoint to generate the ConstructionSite from

        Returns:
            ConstructionSite
        """
        response = agent.client.get(
            f'/systems/{waypoint.system_symbol}/'
            f'waypoints/{waypoint.symbol}/construction'
        )
        data = response.json()['data']
        data['systemSymbol'] = waypoint.system_symbol
        data['symbol'] = waypoint.symbol
        return cls(agent, data)

    @retry()
    def refresh(self):
        """
        Returns a new ConstructionSite class object of the
        current ConstructionSite and its current status

        Returns:
            A new ConstructionSite instance from self
        """
        response = self.agent.client.get(
            f'/systems/{self.system_symbol}/waypoints/'
            f'{self.symbol}/construction'
        )
        data = response.json()['data']
        data['systemSymbol'] = self.system_symbol
        data['symbol'] = self.symbol
        return ConstructionSite(self.agent, data)

    @retry()
    @transit
    @docked
    def supply(
        self,
        *,
        ship,
        trade_symbol,
        units,
    ):
        """
        Supply construction materials to a Waypoint under construction

        Kwargs:
            ship: Ship that contains the construciton materials
                  and is at the Waypoint
            trade_symbol: Symbol of the good to supply
            units: Number of units of the good to supply

        Blocks:
            True: Won't be executed until Ship reaches destination
        """
        units = int(units)
        payload = {
            'shipSymbol': ship.symbol,
            'tradeSymbol': trade_symbol,
            'units': units,
        }
        response = self.agent.client.post(
            f'/systems/{self.system_symbol}/waypoints/{self.symbol}'
            '/construction/supply',
            json=payload
        )
        data = response.json()['data']
        self.update_data_item('materials', data['construction']['materials'])
        self.update_data_item('isComplete', data['construction']['isComplete'])
        ship.update_data_item('cargo', data['cargo'])
        logger.info(
            f'{ship.registration.role}: {ship.symbol} | '
            f'Supplied {units:,} units of '
            f'{trade_symbol} to {self.symbol}'
        )
        return data


# NOTE: These can be convient for isinstance checking


class Planet(utils.AbstractJSONItem):

    def __init__(self, agent, planet):
        self.agent = agent
        self._data = planet


class GasGiant(utils.AbstractJSONItem):

    def __init__(self, agent, gas_giant):
        self.agent = agent
        self._data = gas_giant


class Moon(utils.AbstractJSONItem):

    def __init__(self, agent, moon):
        self.agent = agent
        self._data = moon


class OrbitalStation(utils.AbstractJSONItem):

    def __init__(self, agent, orbital_station):
        self.agent = agent
        self._data = orbital_station


class JumpGate(utils.AbstractJSONItem):

    def __init__(self, agent, jump_gate):
        self.agent = agent
        self._data = jump_gate

    @property
    def data(self):
        response = self.agent.client.get(
            f'/systems/{self.system_symbol}/waypoints/'
            f'{self.symbol}/jump-gate'
        )
        return JumpGateData(self.agent, response.json()['data'])


class JumpGateData(utils.AbstractJSONItem):

    def __init__(self, agent, jump_gate):
        self.agent = agent
        self._data = jump_gate


class AsteroidField(utils.AbstractJSONItem):

    def __init__(self, agent, asteroid_field):
        self.agent = agent
        self._data = asteroid_field


class Asteroid(utils.AbstractJSONItem):

    def __init__(self, agent, asteroid):
        self.agent = agent
        self._data = asteroid


class EngineeredAsteroid(utils.AbstractJSONItem):

    def __init__(self, agent, engineered_asteroid):
        self.agent = agent
        self._data = engineered_asteroid


class AsteroidBase(utils.AbstractJSONItem):

    def __init__(self, agent, asteroid_base):
        self.agent = agent
        self._data = asteroid_base


class Nebula(utils.AbstractJSONItem):

    def __init__(self, agent, nebula):
        self.agent = agent
        self._data = nebula


class DebrisField(utils.AbstractJSONItem):

    def __init__(self, agent, debris_field):
        self.agent = agent
        self._data = debris_field


class GravityWell(utils.AbstractJSONItem):

    def __init__(self, agent, gravity_well):
        self.agent = agent
        self._data = gravity_well


class ArtificialGravityWell(utils.AbstractJSONItem):

    def __init__(self, agent, artificial_gravity_well):
        self.agent = agent
        self._data = artificial_gravity_well


class FuelStation(utils.AbstractJSONItem):

    def __init__(self, agent, fuel_station):
        self.agent = agent
        self._data = fuel_station


def is_uncharted(waypoint):
    return any(
        trait.symbol.upper() == 'UNCHARTED' for trait in waypoint.traits
    )


def best_asteroid(ship, trade_symbol=None, *, contract=None):
    trade_symbols = set()
    if trade_symbol is not None:
        trade_symbols.add(trade_symbol.strip().upper())
    if contract is not None:
        for item in contract.terms.deliver:
            if item.units_required - item.units_fulfilled > 0:
                trade_symbols.add(item.trade_symbol)
    if not trade_symbols:
        trade_symbols = utils.MINABLE_SYMBOLS.union(utils.REFINABLE_SYMBOLS)
    small_surveys = set()
    moderate_surveys = set()
    large_surveys = set()
    for survey_item in ship.waypoints.surveys():
        for survey in survey_item.surveys:
            if any(i.symbol in trade_symbols for i in survey.deposits):
                if survey.size == 'SMALL':
                    small_surveys.add(survey.symbol)
                elif survey.size == 'MODERATE':
                    moderate_surveys.add(survey.symbol)
                elif survey.size == 'LARGE':
                    large_surveys.add(survey.symbol)
    if large_surveys:
        large_waypoints = (
            ship.waypoints.get(waypoint_symbol=i) for i in large_surveys
        )
        return min(large_waypoints, key=lambda x: ship.distance(x))
    if moderate_surveys:
        moderate_waypoints = (
            ship.waypoints.get(waypoint_symbol=i) for i in moderate_surveys
        )
        return min(moderate_waypoints, key=lambda x: ship.distance(x))
    if small_surveys:
        small_waypoints = (
            ship.waypoints.get(waypoint_symbol=i) for i in small_surveys
        )
        return min(small_waypoints, key=lambda x: ship.distance(x))


@functools.lru_cache
def class_factory(cls_name):
    cls_name = utils.camel_case(cls_name.capitalize().rstrip('s'))
    try:
        return globals()[cls_name]
    except KeyError:
        return type(cls_name, (utils.BaseItem,), {})



import itertools
import logging
import os
import re
import time


logger = logging.getLogger(__name__)


try:
    from rich import print
    from snisp.utils import RichHighlighter as LogHighlighter
except ModuleNotFoundError:
    from snisp.utils import DummyHighlighter as LogHighlighter
    logger.warning('"rich" not installed. Defaulting to standard print.')


def tail(
    log_file,
    *,
    color_regexes=None,
    match_regexes=None,
    filter_regexes=None,
    show_httpx=False,
    sleep=1
):
    highlighter = LogHighlighter()
    if color_regexes is not None:
        for regex, colors in color_regexes:
            highlighter.regexes.append((regex, colors))
    if match_regexes is None:
        match_regexes = []
    else:
        match_regexes = [re.compile(r) for r in match_regexes]
    if filter_regexes is None:
        filter_regexes = []
    else:
        filter_regexes = [re.compile(r) for r in filter_regexes]
    if not os.path.isfile(log_file):
        logger.error(f'Logfile {log_file} does not exist. Cannot tail.')
        return
    try:
        _ = open(log_file, 'rb')
    except PermissionError:
        logger.error(
            f'User does not have permission to open {log_file}. Cannot tail.'
        )
        return
    with open(log_file, 'rb') as f:
        f.seek(0, os.SEEK_END)
        try:
            while True:
                last_write_pos = f.tell()
                if data := f.readline():
                    line = data.decode('utf8', errors='ignore').strip()
                    if not show_httpx and ' httpx |' in line:
                        continue
                    if any(r.search(line) for r in filter_regexes):
                        continue
                    if match_regexes:
                        if not any(r.search(line) for r in match_regexes):
                            continue
                    print(highlighter(line))
                else:
                    f.seek(last_write_pos)
                    time.sleep(sleep)
        except KeyboardInterrupt:
            pass


def regex_colors(arg):
    try:
        args = arg.split(',')
        regex = re.compile(rf'{args[0]}')
        colors = itertools.cycle(args[1:])
        return regex, colors
    except Exception as e:
        raise argparse.ArgumentTypeError(
            f'{str(e)}: "regexes" expects an input of "regex,color", etc.'
        )


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        'log_file', help='Full path for the log file to be tailed.'
    )
    parser.add_argument(
        '-p',
        '--patterns',
        nargs='+',
        type=regex_colors,
        help='Regex patterns to pass to Rich. Requires "rich" to be '
             r'installed.\ne.g.,--paterns  "\d+, blue on white,red on green" '
             r'"\w, red,green"\nThe pattern is "{regex}, {colors},..." where '
             '{colors} will be cycled over all matches.'
    )
    parser.add_argument(
        '-m',
        '--matches',
        nargs='+',
        help='List of regexes to filter the logs. Only logs that match any of '
             'the conditions in the regexes will be shown.'
    )
    parser.add_argument(
        '-f',
        '--filters',
        nargs='+',
        help='List of regexes to filter the logs. Logs that match any of '
             'the conditions in the regexes will be NOT shown.'
    )
    parser.add_argument(
        '--show-httpx',
        default=False,
        action='store_true',
        help='Flag to enable showing logs created '
             'by "httpx" in the tail output.'
    )
    args = parser.parse_args()

    tail(
        args.log_file,
        color_regexes=args.patterns,
        match_regexes=args.matches,
        filter_regexes=args.filters,
        show_httpx=args.show_httpx,
    )



import itertools
import queue
import threading
import time

from snisp.agent import Agent


"""
This recipe shows how to more efficiently strip all of the minable Asteroids
in the Agent's current System. By using a Queue of Asteroid Symbols, the recipe
demonstrates how Waypoints can be easily serialized. The shared Queue between
threads is also a more efficient means of extraction, compared to the
default example in the README.

The recipe can be run in it's own thread and it requires no monitoring,
which is ideal for running as a sub-script for a larger game.

Each new iteration of agent.fleet will return the most up-to-date
list of available Ships. Meaning, if you purchase additional Ships
in a seperate thread in the time between iterations and those Ships can_mine,
the Ships will automatically show up here.
"""


def main(*, agent_symbol, faction='', email='', token=''):
    # Get your agent
    agent = Agent(
        symbol=agent_symbol, faction=faction, email=email, token=token
    )

    # Fire off a thread
    # The thread will run until all Asteroids have been stripped
    # or the program exits
    t = threading.Thread(target=run, args=(agent,), daemon=True)
    t.start()


def run(agent):
    # Get your command ship
    command_ship = next(iter(agent.fleet))

    # Get all of the mineable Asteroids Symbols in your Command Ship's System
    # Sets automatically prevent duplicates
    asteroid_symbols = {
        i.symbol for i in
        itertools.chain(
            command_ship.waypoints.asteroids(),
            command_ship.waypoints.engineered_asteroids(),
        )
    }
    # Build a Queue with all of the Symbols
    # You could just build the Queue with the Asteroids directly but one of
    # the intents of this recipe is to show how Waypoints can go from
    # symbol --> Waypoint, which makes serializing data a breeze
    asteroid_queue = queue.Queue()
    for symbol in asteroid_symbols:
        asteroid_queue.put(symbol)

    already_mining_ships = set()

    wait = 60 * 15  # Fifteen minutes between loops
    while not asteroid_queue.empty():
        for ship in agent.fleet:
            if ship.symbol not in already_mining_ships:
                if ship.can_mine:
                    t = threading.Thread(
                        target=extract,
                        args=(ship, asteroid_queue),
                        daemon=True
                    )
                    t.start()
                    already_mining_ships.add(ship.symbol)
        for _ in range(wait):
            time.sleep(1)
            # Check the queue once a second
            if asteroid_queue.empty():
                # All Asteroids are being extracted! Need to break out of the
                # loop and wait until they're completed
                break
    # Block until all currently working threads have had a chance to strip
    # their Asteroids and mark their queue task done
    asteroid_queue.join()


def extract(ship, asteroid_queue):
    try:
        # Attempt to get an Asteroid symbol
        asteroid_symbol = asteroid_queue.get_nowait()
    except queue.Empty:
        # No more asteroids. Exit
        return
    else:
        # Get the Asteroid Waypoint from the Asteroid symbol
        asteroid = ship.waypoints.get(waypoint_symbol=asteroid_symbol)

    while True:
        # Navigate to the Asteroid
        # Autopilot blocks until you're at the destination
        ship.autopilot(asteroid)
        stripped = False
        # Extract until cargo is full
        # ship.extract automatically updates the ship's cargo
        while ship.cargo.units < ship.cargo.capacity:
            extraction = ship.extract()
            if extraction.units == 0:
                # Asteroid is stripped! Nothing else to see here
                stripped = True
                break
        # Sell off whatever is in the Ship's cargo at the closest Markets
        # Jettisons anything that can't be sold
        ship.sell_off_cargo()
        if stripped:
            # Mark the Asteroid as completed
            asteroid_queue.task_done()
            try:
                # Attempt to get the next Asteroid Symbol
                asteroid_symbol = asteroid_queue.get_nowait()
            except queue.Empty:
                # No more asteroids. Exit
                return
            else:
                # Get the Asteroid Waypoint from the Asteroid symbol
                asteroid = ship.waypoints.get(waypoint_symbol=asteroid_symbol)



import dateutil
import json
import logging
import os
import sqlite3

from datetime import datetime, timezone
from threading import Lock

# :memory: has rough edges
DATABASE = os.path.join(
    os.path.abspath(os.path.dirname(__file__)), 'data', 'cache.db'
)
DATABSE_LOCK = Lock()
logger = logging.getLogger(__name__)


def get_waypoints(
        *,
        system,
        page_limit,
        page=0,
        traits=None,
        types=None,
        cache_mins=15,
):
    con = sqlite3.connect(DATABASE)
    cur = con.cursor()
    params = {
        'system': system,
        'page_limit': int(page_limit),
        'page': int(page),
        'type': types if types is not None else '',
        'traits': traits if traits is not None else '',
    }
    keys = params.keys()  # I know it's ordered now but still
    where_string = ' AND '.join(f'{k} = (?)' for k in keys)
    query = cur.execute(f'''
        SELECT
            data, last_updated
        FROM
            waypoints
        WHERE
            {where_string}
        ''', tuple(params[k] for k in keys),
    )
    result = query.fetchone()
    con.close()
    if result:
        if last_updated := result[1]:  # pragma: no cover
            last_updated = dateutil.parser.parse(last_updated)
            delta = datetime.now(timezone.utc) - last_updated
            if delta.seconds and (delta.seconds / 60) >= cache_mins:
                return
        return json.loads(result[0])


def insert_waypoints(
    data,
    *,
    system,
    page_limit,
    page=0,
    types=None,
    traits=None,
):
    con = sqlite3.connect(DATABASE)
    cur = con.cursor()
    params = {
        'system': system,
        'page_limit': int(page_limit),
        'page': int(page),
        'last_updated': datetime.now(timezone.utc).isoformat(),
        'type': types if types is not None else '',
        'traits': traits if traits is not None else '',
    }
    keys = params.keys()  # I know it's ordered now but still
    insert_string = ', '.join(keys)
    values_string = ', '.join('?' for _ in range(len(keys)))
    update_values = [params[k] for k in keys]
    update_values.insert(0, json.dumps(data))

    # INSERT OR REPLACE/REPLACE INTO wasn't working reliably
    where_string = ' AND '.join(
        f'{k} = (?)' for k in keys if k != 'last_updated'
    )
    _ = cur.execute(f'''
        DELETE FROM
            waypoints
        WHERE
            ({where_string})
        ''', tuple(params[k] for k in keys if k != 'last_updated'))

    _ = cur.execute(f'''
        INSERT INTO
            waypoints (data, {insert_string})
        VALUES
            (json(?), {values_string})
        ''', update_values)
    result = False
    if cur.rowcount:
        con.commit()
        result = True
    else:  # pragma: no cover
        logger.warning(
            f'Failed to insert data for waypoints at page {page:,}'
        )
        con.rollback()
    con.close()
    return result


def setup():
    if not os.path.isfile(DATABASE):  # pragma: no cover
        os.makedirs(os.path.dirname(DATABASE), exist_ok=True)
    con = sqlite3.connect(DATABASE)
    cur = con.cursor()
    _ = con.execute('''
        CREATE TABLE IF NOT EXISTS waypoints (
            pk INTEGER PRIMARY KEY ASC,
            system TEXT NOT NULL,
            page_limit INTEGER NOT NULL,
            page INTEGER NOT NULL,
            traits TEXT,
            type TEXT,
            last_updated TEXT,
            data TEXT NOT NULL
        );''')
    _ = con.commit()
    _ = cur.execute('PRAGMA journal_mode=WAL;')
    _ = con.commit()
    _ = cur.execute('DELETE FROM waypoints')
    _ = con.commit()
    con.close()



import logging

from snisp import exceptions, utils
from snisp.decorators import cooldown, in_orbit, retry, transit


logger = logging.getLogger(__name__)


class Systems:

    def __init__(self, agent):
        self.agent = agent

    def __repr__(self):
        return f'{self.__class__.__name__}({self.agent!r})'

    def __iter__(self):
        """
        Iterates over the System's in the Agent's current universe

        Yields:
            StarSystem
        """
        page = 1
        response = self.get_page(page=page)
        while data := response.json()['data']:
            for system in data:
                yield StarSystem(self.agent, system)
            page += 1
            response = self.get_page(page=page)

    def find(self, **filters):
        """
        Find the first instance of the StarSystems which match the filters

        Kwargs:
            **filters: Top-level key, value pairs used to filter the Systems

        Returns:
            StarSystem if **filters leads to a match; else, None
        """
        try:
            return next(self.find_all(**filters))
        except StopIteration:
            logger.info(f'No system found with {filters!r}')

    def find_all(self, **filters):
        """
        Find the all instances of the StarSystems which match the filters

        Kwargs:
            **filters: Top-level key, value pairs used to filter the Systems

        Yields:
            StarSystem
        """
        if _type := filters.get('type'):
            _type = _type.upper().strip()
            if _type not in utils.SYSTEMS_TYPES:
                raise exceptions.SpaceAttributeError(
                    f'{_type} is not an acceptable System type.'
                )
            filters['type'] = _type
        filters = {utils.camel_case(k): v for k, v in filters.items()}
        for system in self:
            system_dict = system.to_dict()
            if all(system_dict[k] == v for k, v in filters.items()):
                yield system

    @retry()
    def get_page(self, page=1, limit=20):
        params = {'limit': int(limit), 'page': int(page)}
        return self.agent.client.get('/systems', params=params)


class System:

    def __init__(self, agent):
        self.agent = agent

    def __repr__(self):
        return f'{self.__class__.__name__}({self.agent!r})'

    @retry()
    def __call__(self, system_symbol):
        """Returns the System for system_symbol"""
        response = self.agent.client.get(f'/systems/{system_symbol}')
        return StarSystem(self.agent, response.json()['data'])

    @retry()
    @transit
    @in_orbit
    @cooldown
    def scan(self, ship):
        """
        Scans the System in the ship's current System

        Args:
            ship: Ship to use. All snisp.fleet.Ship objects will have this
                  automatically monkey-patched in.
        Blocks:
            True: Won't be executed until Ship reaches destination and/or
                  the cooldown period has passed

        Yields:
            StarSystem
        """
        response = self.agent.client.post(
            f'/my/ships/{ship.symbol}/scan/systems'
        )
        data = response.json()['data']
        ship.update_data_item('cooldown', data['cooldown'])
        for _system in data['systems']:
            yield StarSystem(self.agent, _system)


class StarSystem(utils.AbstractJSONItem):

    def __init__(self, agent, system):
        self.agent = agent
        self._data = system


class Location(utils.AbstractJSONItem):

    """
    Location is used throughout as a convenience class for normalizing
    sector, system, and waypoint symbols
    """

    def __init__(self, agent, location):
        self.agent = agent
        self._data = location

    @property
    def sector(self):
        """Property that returns the Location's sector"""
        if hq := self.headquarters:
            return hq.split('-')[0]
        return self.waypoint.split('-')[0]

    @property
    def system(self):
        """Property that returns the Location's system"""
        if hq := self.headquarters:
            return '-'.join(hq.split('-')[0:2])
        return '-'.join(self.waypoint.split('-')[0:2])

    @property
    def waypoint(self):
        """Property that returns the Location's waypoint"""
        if hq := self.headquarters:
            return hq
        if nav := self.nav:
            if wp := nav.waypoint_symbol:
                return wp
        return self.symbol



from snisp import exceptions, utils
from snisp.decorators import retry


class Factions:

    """Factions found in Agent's universe"""

    def __init__(self, agent):
        self.agent = agent

    def __repr__(self):
        return f'{self.__class__.__name__}({self.agent!r})'

    def __iter__(self):
        """
        Iterates over the Factions in the universe

        Yields:
            Faction
        """
        page = 1
        response = self.get_page(page=page)
        while data := response.json()['data']:
            for faction in data:
                yield Faction(self.agent, faction)
            page += 1
            response = self.get_page(page=page)

    @retry()
    def __call__(self, faction_symbol):
        """
        Returns a Faction associated with the faction symbol

        Args:
            faction_symbol: The Faction's symbol. See snip.utils.FACITON_SYMBOL
                            for support Faction symbols

        Returns:
            Faction
        """
        faction_symbol = faction_symbol.upper().strip()
        if faction_symbol not in utils.FACTION_SYMBOL:
            raise exceptions.WaypointNoFactionError(
                f'{faction_symbol} is not an acceptable Faction symbol. '
                'See snisp.utils.FACTION_SYMBOLS for acceptable symbols.'
            )
        response = self.agent.client.get(f'/factions/{faction_symbol}')
        return Faction(self.agent, response.json()['data'])

    @retry()
    def get_page(self, page=1, limit=20):
        params = {'limit': int(limit), 'page': int(page)}
        return self.agent.client.get('/factions', params=params)


class Faction(utils.AbstractJSONItem):

    def __init__(self, agent, faction):
        self.agent = agent
        self._data = faction



import dateutil
import logging

from datetime import datetime, timezone

from snisp import utils
from snisp.decorators import docked, retry, transit


logger = logging.getLogger(__name__)


class Contracts:

    def __init__(self, agent):
        self.agent = agent
        self.last_contract_page = 1

    def __iter__(self):
        """
        Iterates over the current Agent's Contracts

        Yields:
            Contract
        """
        page = 1
        response = self.get_page(page=page)
        while data := response.json()['data']:
            page += 1
            for contract in data:
                yield Contract(self.agent, contract)
            response = self.get_page(page=page)

    @retry()
    def __call__(self, contract_id):
        """Returns the Contract assoiated with the contract ID

        Args:
            contract_id: The Contract's ID to lookup

        Returns:
            Contract
        """
        response = self.agent.client.get(f'/my/contracts/{contract_id}')
        return Contract(self.agent, response.json()['data'])

    @property
    def current(self):
        """Returns the current Contract"""
        with self.agent.lock:
            response = self.get_page(page=self.last_contract_page)
            while data := response.json()['data']:
                self.last_contract_page += 1
                for _contract in data:
                    contract = Contract(self.agent, _contract)
                response = self.get_page(page=self.last_contract_page)
            self.last_contract_page -= 1
        return contract

    @retry()
    def get_page(self, page=1, limit=20):
        params = {'limit': int(limit), 'page': int(page)}
        return self.agent.client.get('/my/contracts', params=params)


class Contract(utils.AbstractJSONItem):

    def __init__(self, agent, contract):
        self.agent = agent
        self._data = contract

    @property
    def expired(self):
        """Property for if the Contract has already expired

        Returns:
            bool
        """
        expiration = dateutil.parser.parse(self.terms.deadline)
        return datetime.now(timezone.utc) >= expiration

    @property
    def extractable(self):
        """Property for if any remaining items to deliver for the Contract
        can be extracted at an Asteroid

        Returns:
            bool
        """
        for term in self.terms.deliver:
            if term.units_required - term.units_fulfilled > 0:
                if term.trade_symbol in utils.MINABLE_SYMBOLS:
                    return True
        return False

    @property
    def siphonable(self):
        """Property for if any remaining items to deliver for the Contract
        can be siphoned at a Gas Giant

        Returns:
            bool
        """
        for term in self.terms.deliver:
            if term.units_required - term.units_fulfilled > 0:
                if term.trade_symbol in utils.SIPHONABLE_SYMBOLS:
                    return True
        return False

    @retry()
    def accept(self):
        """
        Accepts the Contract. Can be called multiple times without an issue
        """
        if not self.accepted:
            self.agent.client.post(
                f'/my/contracts/{self.id}/accept'
            )
            self.update_data_item('accepted', True)
            logger.info(f'Contract {self.id} accepted')

    @retry()
    @transit
    @docked
    def deliver(self, ship, trade_symbol, max_units=0):
        """
        Deliver items to the Contract.

        Ship must be located at the Contracts respective contract.deliver
        "destination_symbol"

        Args:
            ship: The Ship that contains the items to deliver
            trade_symbol: The symbol of the item to deliver

        Kwargs:
            max_units: Number of units to deliver, if > 1. Otherwise,
                       either the number of remaining units to be delivered
                       or the number of units in the Ship's cargo, whatever
                       is fewer

        Blocks:
            True: Won't be executed until Ship reaches destination
        """
        trade_symbol = trade_symbol.strip().upper()
        cargo_units = next(
            (
                int(i.units) for i in ship.cargo.inventory
                if i.symbol == trade_symbol
            ), None
        )
        if cargo_units is None:
            logger.warning(
                f'{ship.registration.role}: {ship.symbol} | '
                f'Attempted to deliver {trade_symbol} '
                f'to Contract {self.id}. '
                f'No {trade_symbol} units found in cargo'
            )
            return
        contract_units = next(
            (
                i.units_required - i.units_fulfilled
                for i in self.terms.deliver
                if i.trade_symbol == trade_symbol
            ), None
        )
        if contract_units is None:
            logger.warning(
                f'{ship.registration.role}: {ship.symbol} | '
                f'Attempted to deliver {trade_symbol} '
                f'to Contract {self.id}. '
                f'Contract does not required {trade_symbol}'
            )
            return
        max_units = max_units if max_units > 0 else cargo_units
        units = min(contract_units, cargo_units, max_units)
        payload = {
            'shipSymbol': ship.symbol,
            'tradeSymbol': trade_symbol,
            'units': units,
        }
        response = self.agent.client.post(
            f'/my/contracts/{self.id}/deliver', json=payload
        )
        data = response.json()['data']
        self.update_data_item('terms', data['contract']['terms'])
        ship.update_data_item('cargo', data['cargo'])
        logger.info(
            f'{ship.registration.role}: {ship.symbol} | '
            f'Delivered {units:,} units of '
            f'{trade_symbol} to CONTRACT {self.id}'
        )
        return data

    @retry()
    def fulfill(self):
        """
        Fulfills the Contract. Can be called multiple times without an issue
        """
        if not self.fulfilled:
            response = self.agent.client.post(
                f'/my/contracts/{self.id}/fulfill'
            )
            data = response.json()['data']
            self.update_data_item('fulfilled', data['contract']['fulfilled'])
            logger.info(f'Contract {self.id} fulfilled!')
            return data

    @retry()
    def refresh(self):
        """Returns a new Contract class object of the current Contract

        Returns:
            A new Contract instance from self
        """
        response = self.agent.client.get(f'/my/contracts/{self.id}')
        return Contract(self.agent, response.json()['data'])



import httpx
import json
import pytest
import os

import snisp

from . import DATA_DIR, attribute_test


class TestAgent:

    agent = snisp.agent.Agent(symbol='testing', faction='testing')

    def test_agent(self, respx_mock):
        assert self.agent.symbol == 'TESTING'
        assert self.agent.faction == 'TESTING'
        assert self.agent.email == 'TESTING'

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_data(self, respx_mock):
        json_data = json.load(
            open(os.path.join(DATA_DIR, 'agent_data.json'), encoding='utf8')
        )
        respx_mock.get('/my/agent').mock(
            return_value=httpx.Response(200, json=json_data)
        )
        response = self.agent.data
        assert response == snisp.agent.PlayerData(
            self.agent, json_data['data']
        )
        attribute_test(response, json_data['data'])



import httpx
import pytest

import snisp


class TestSpaceClient:

    token = snisp.client.load_user(symbol='testing', faction='testing')
    client = snisp.client.SpaceClient(token=token)

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_init(self, respx_mock):
        assert str(self.client.base_url) == 'https://api.spacetraders.io/v2/'

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_closing(self, respx_mock):
        client = snisp.client.SpaceClient(token=self.token)
        assert not client.is_closed

        client = snisp.client.SpaceClient(token=self.token)
        assert not client.is_closed
        client.cleanup()
        assert client.is_closed

        client = snisp.client.SpaceClient(token=self.token)
        assert not client.is_closed
        client.__del__()  # Not great. Just used to shut up pytest
        assert client.is_closed

        client = snisp.client.SpaceClient(token=self.token)
        assert not client.is_closed
        snisp.client.cleanup_client(client)
        assert client.is_closed
        snisp.client.cleanup_client('fake_client')

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_unused_options(self, respx_mock):
        # Now with real fake tests!
        # (this actually exposed what would have been a bug in
        # one of the unused paths) :/
        respx_mock.delete('').mock(
            httpx.Response(200, json={'data': 'delete'})
        )
        response = self.client.delete('/')
        assert response.json()['data'] == 'delete'
        respx_mock.options('').mock(
            httpx.Response(200, json={'data': 'options'})
        )
        response = self.client.options('')
        assert response.json()['data'] == 'options'
        respx_mock.put('/').mock(
            httpx.Response(200, json={'data': 'put'})
        )
        response = self.client.put('')
        assert response.json()['data'] == 'put'



import os
import logging
import threading

from snisp import agent, cache, database, utils  # noqa: F401


logging.basicConfig(
    format='%(asctime)s %(levelname)s %(name)s | %(message)s',
    datefmt='%m/%d/%Y %I:%M:%S %p',
    encoding='utf-8',
    filename=os.path.join(
        os.path.abspath(os.path.dirname(__file__)), 'snisp.log'
    ),
    filemode='w',
    level=logging.INFO,
)


threading.excepthook = utils.thread_exception_hook_logger



import copy
import logging
import math
import threading
import traceback

from collections.abc import Iterable

try:  # pragma: no cover
    from rich.highlighter import Highlighter
except ModuleNotFoundError:  # pragma: no cover
    class Highlighter:
        pass


ATTRIBUTE_LOCK = threading.Lock()
logger = logging.getLogger(__name__)


class AbstractJSONItem:

    """Allows for snake or camel case accessing of JSON data
    as class attributes.
    """

    def __dir__(self):
        output = super().__dir__()
        output.extend(
            snake_case(i) for i in self._data.keys()
            if snake_case(i) not in output
        )
        return output

    def __repr__(self):
        return f'{self.__class__.__name__}({self.to_dict()!r})'

    def __eq__(self, other):
        return type(other) == type(self) and other.to_dict() == self.to_dict()

    def __len__(self):
        return len(self._data)

    def __getattribute__(self, name):
        # Slow but fight me
        if name == 'agent':
            return super().__getattribute__('agent')
        data_dict = super().__getattribute__('_data')
        if name == '_data':
            return data_dict
        try:
            if name in data_dict.keys():
                value = super().__getattribute__(name)
                if value != data_dict[name]:
                    super().__setattr__(name, data_dict[name])
                return data_dict[name]
            _name = camel_case(name)
            if _name in data_dict.keys():
                value = super().__getattribute__(_name)
                if value != data_dict[_name]:  # pragma: no cover
                    # Probably not possible to get here since update_data_item
                    # also makes a call to __setattr__
                    # buy why not?
                    super().__setattr__(_name, data_dict[_name])
                return data_dict[_name]
            return super().__getattribute__(name)
        except AttributeError:
            if name in data_dict.keys():
                if isinstance(data_dict[name], dict):
                    data = type(
                        camel_case(name.capitalize()).rstrip('s'),
                        (BaseJSONItem,),
                        {}
                    )(data_dict[name])
                    data_dict[name] = data
                    super().__setattr__(name, data)
                    return data
                elif is_list_like(data_dict[name]):
                    class_factory = type(
                        camel_case(name.capitalize()).rstrip('s'),
                        (BaseJSONItem,),
                        {}
                    )
                    data = []
                    for index, item in enumerate(data_dict[name]):
                        if type(item) == str:
                            data.append(item)
                        else:
                            base_item = class_factory(item)
                            data_dict[name][index] = base_item
                            data.append(base_item)
                    super().__setattr__(name, data)
                else:
                    super().__setattr__(name, data_dict[name])
                return data_dict[name]
            name = camel_case(name)
            if name in data_dict.keys():
                if isinstance(data_dict[name], dict):  # pragma: no cover
                    data = type(
                        ''.join([name[0].upper(), name[1:]]).rstrip('s'),
                        (BaseJSONItem,),
                        {}
                    )(data_dict[name])
                    data_dict[name] = data
                    super().__setattr__(name, data)
                    return data
                elif is_list_like(data_dict[name]):
                    class_factory = type(
                        ''.join([name[0].upper(), name[1:]]).rstrip('s'),
                        (BaseJSONItem,),
                        {}
                    )
                    data = []
                    for index, item in enumerate(data_dict[name]):
                        if type(item) == str:  # pragma: no cover
                            data.append(item)
                        else:
                            base_item = class_factory(item)
                            data_dict[name][index] = base_item
                            data.append(base_item)
                    super().__setattr__(name, data)
                else:
                    super().__setattr__(name, data_dict[name])
                return data_dict[name]

    def __setattr__(self, name, value):
        if name == '_data' or name == 'agent':
            return super().__setattr__(name, value)
        raise AttributeError(
            f'{self.__class__.__name__} does not support attribute assignment'
        )

    def to_dict(self):
        output = {}
        for k, v in self._data.items():
            if issubclass(type(v), AbstractJSONItem):
                output[k] = v.to_dict()
            elif is_list_like(v) and not isinstance(v, dict):
                output[k] = []
                for item in v:
                    if issubclass(type(item), AbstractJSONItem):
                        output[k].append(item.to_dict())
                    else:
                        output[k].append(copy.copy(item))
            else:
                output[k] = copy.copy(v)
        return output

    def update_data_item(self, name, value):
        """Updates top-level items found in the `_data` dict.
        Not intended to be used by a user."""
        with ATTRIBUTE_LOCK:
            if name not in self._data.keys():
                raise AttributeError(
                    f'{name!r} not found in {self.__class__.__name__}'
                )
            if isinstance(value, dict):
                value = type(
                    camel_case(name.capitalize()).rstrip('s'),
                    (BaseJSONItem,),
                    {}
                )(value)
            elif is_list_like(value):
                class_factory = type(
                    camel_case(name.capitalize()).rstrip('s'),
                    (BaseJSONItem,),
                    {}
                )
                data = []
                for index, item in enumerate(value):
                    if type(item) == str:  # pragma: no cover
                        data.append(item)
                    else:
                        data.append(class_factory(item))
                value = data
            self._data[name] = value
            super().__setattr__(name, value)


class BaseItem(AbstractJSONItem):

    def __init__(self, agent, data):
        self.agent = agent
        self._data = data


class BaseJSONItem(AbstractJSONItem):

    def __init__(self, data):
        self._data = data


class DummyHighlighter:  # pragma: no cover

    regexes = []

    def __call__(self, text):
        return text


class RichHighlighter(Highlighter):  # pragma: no cover

    regexes = []

    def highlight(self, text):
        plain_text = text.plain
        for regex, colors in self.regexes:
            for match in regex.finditer(plain_text):
                if (num_groups := len(match.groups())):
                    for group in range(1, num_groups+1):
                        text.stylize(
                            next(colors),
                            match.span(group)[0],
                            match.span(group)[1]
                        )
                else:
                    text.stylize(
                        next(colors), match.span(0)[0], match.span(0)[1]
                    )


def is_list_like(value):
    if isinstance(value, Iterable):
        if not isinstance(value, str):
            return True
    return False


def snake_case(name):
    if name.startswith('__') and name.endswith('__'):  # pragma: no cover
        return name
    output = ''
    for char in name:
        if char.isupper():
            output += f'_{char.lower()}'
        else:
            output += char
    return output


def camel_case(name):
    if name.startswith('__') and name.endswith('__'):
        return name
    upper_next = False
    output = ''
    for char in name:
        if char == '_':
            upper_next = True
        else:
            if upper_next:
                char = char.upper()
                upper_next = False
            output += char
    return output


# Helpers

def ilen(*iterables, start=0):
    for iterable in iterables:
        for _ in iterable:
            start += 1
    return start


def calculate_distance(ship, dest):
    return math.sqrt(
        pow(ship.nav.route.destination.x - dest.x, 2) +
        pow(ship.nav.route.destination.y - dest.y, 2)
    )


def calculate_waypoint_distance(origin, dest):
    return math.sqrt(pow(origin.x - dest.x, 2) + pow(origin.y - dest.y, 2))


def a_ship_at_location(agent, symbol):
    return any(
        s.nav.waypoint_symbol == symbol
        for s in agent.fleet if s.nav.status != 'IN_TRANSIT'
    )


def thread_exception_hook_logger(err):  # pragma: no cover
    logger.warning(f'ThreadExcetion: {err!r}')
    logger.warning(traceback.extract_tb(err.exc_traceback))


def thread_exception_hook_raiser(err):  # pragma: no cover
    print(traceback.extract_tb(err.exc_traceback))
    raise Exception(err)


# AKA enums

SYSTEMS_TYPES = frozenset(
    [
        'BLACK_HOLE',
        'BLUE_STAR',
        'HYPERGIANT',
        'NEBULA',
        'NEUTRON_STAR',
        'ORANGE_STAR',
        'RED_STAR',
        'UNSTABLE',
        'YOUNG_STAR',
        'WHITE_DWARF',
    ]
)


WAYPOINT_TRAITS = frozenset(
    [
        'ASH_CLOUDS',
        'BARREN',
        'BLACK_MARKET',
        'BREATHABLE_ATMOSPHERE',
        'BUREAUCRATIC',
        'CANYONS',
        'COMMON_METAL_DEPOSITS',
        'CORROSIVE_ATMOSPHERE',
        'CORRUPT',
        'CRUSHING_GRAVITY',
        'DEBRIS_CLUSTER',
        'DEEP_CRATERS',
        'DIVERSE_LIFE',
        'DRY_SEABEDS',
        'EXPLORATION_OUTPOST',
        'EXPLOSIVE_GASES',
        'EXTREME_PRESSURE',
        'EXTREME_TEMPERATURES',
        'FOSSILS',
        'FROZEN',
        'HIGH_TECH',
        'HOLLOWED_INTERIOR',
        'ICE_CRYSTALS',
        'INDUSTRIAL',
        'JOVIAN',
        'JUNGLE',
        'MAGMA_SEAS',
        'MARKETPLACE',
        'MEGA_STRUCTURES',
        'METHANE_POOLS',
        'MICRO_GRAVITY_ANOMALIES',
        'MILITARY_BASE',
        'MINERAL_DEPOSITS',
        'MUTATED_FLORA',
        'OCEAN',
        'OUTPOST',
        'OVERCROWDED',
        'PERPETUAL_DAYLIGHT',
        'PERPETUAL_OVERCAST',
        'PIRATE_BASE',
        'PRECIOUS_METAL_DEPOSITS',
        'RADIOACTIVE',
        'RARE_METAL_DEPOSITS',
        'RESEARCH_FACILITY',
        'ROCKY',
        'SALT_FLATS',
        'SCARCE_LIFE',
        'SCATTERED_SETTLEMENTS',
        'SHALLOW_CRATERS',
        'SHIPYARD',
        'SPRAWLING_CITIES',
        'STRIPPED',
        'STRONG_GRAVITY',
        'STRONG_MAGNETOSPHERE',
        'SUPERVOLCANOES',
        'SURVEILLANCE_OUTPOST',
        'SWAMP',
        'TEMPERATE',
        'TERRAFORMED',
        'THIN_ATMOSPHERE',
        'TOXIC_ATMOSPHERE',
        'TRADING_HUB',
        'UNCHARTED',
        'UNDER_CONSTRUCTION',
        'UNSTABLE_COMPOSITION',
        'VAST_RUINS',
        'VIBRANT_AURORAS',
        'VOLCANIC',
        'WEAK_GRAVITY',
    ]
)


WAYPOINT_TYPES = frozenset(
    [
        'ARTIFICIAL_GRAVITY_WELL',
        'ASTEROID',
        'ASTEROID_BASE',
        'ASTEROID_FIELD',
        'DEBRIS_FIELD',
        'ENGINEERED_ASTEROID',
        'FUEL_STATION',
        'GAS_GIANT',
        'GRAVITY_WELL',
        'JUMP_GATE',
        'MOON',
        'NEBULA',
        'ORBITAL_STATION',
        'PLANET',
    ]
)


SHIP_TYPES = frozenset(
    [
        'SHIP_COMMAND_FRIGATE',
        'SHIP_EXPLORER',
        'SHIP_HEAVY_FREIGHTER',
        'SHIP_INTERCEPTOR',
        'SHIP_LIGHT_HAULER',
        'SHIP_LIGHT_SHUTTLE',
        'SHIP_MINING_DRONE',
        'SHIP_ORE_HOUND',
        'SHIP_PROBE',
        'SHIP_REFINING_FREIGHTER',
        'SHIP_SIPHON_DRONE',
        'SHIP_SURVEYOR',
    ]
)

GOODS_TYPES = frozenset(
    [
        'PRECIOUS_STONES',
        'QUARTZ_SAND',
        'SILICON_CRYSTALS',
        'AMMONIA_ICE',
        'LIQUID_HYDROGEN',
        'LIQUID_NITROGEN',
        'ICE_WATER',
        'EXOTIC_MATTER',
        'ADVANCED_CIRCUITRY',
        'GRAVITON_EMITTERS',
        'IRON',
        'IRON_ORE',
        'COPPER',
        'COPPER_ORE',
        'ALUMINUM',
        'ALUMINUM_ORE',
        'SILVER',
        'SILVER_ORE',
        'GOLD',
        'GOLD_ORE',
        'PLATINUM',
        'PLATINUM_ORE',
        'DIAMONDS',
        'URANITE',
        'URANITE_ORE',
        'MERITIUM',
        'MERITIUM_ORE',
        'HYDROCARBON',
        'ANTIMATTER',
        'FAB_MATS',
        'FERTILIZERS',
        'FABRICS',
        'FOOD',
        'JEWELRY',
        'MACHINERY',
        'FIREARMS',
        'ASSAULT_RIFLES',
        'MILITARY_EQUIPMENT',
        'EXPLOSIVES',
        'LAB_INSTRUMENTS',
        'AMMUNITION',
        'ELECTRONICS',
        'SHIP_PLATING',
        'SHIP_PARTS',
        'EQUIPMENT',
        'FUEL',
        'MEDICINE',
        'DRUGS',
        'CLOTHING',
        'MICROPROCESSORS',
        'PLASTICS',
        'POLYNUCLEOTIDES',
        'BIOCOMPOSITES',
        'QUANTUM_STABILIZERS',
        'NANOBOTS',
        'AI_MAINFRAMES',
        'QUANTUM_DRIVES',
        'ROBOTIC_DRONES',
        'CYBER_IMPLANTS',
        'GENE_THERAPEUTICS',
        'NEURAL_CHIPS',
        'MOOD_REGULATORS',
        'VIRAL_AGENTS',
        'MICRO_FUSION_GENERATORS',
        'SUPERGRAINS',
        'LASER_RIFLES',
        'HOLOGRAPHICS',
        'SHIP_SALVAGE',
        'RELIC_TECH',
        'NOVEL_LIFEFORMS',
        'BOTANICAL_SPECIMENS',
        'CULTURAL_ARTIFACTS',
        'FRAME_PROBE',
        'FRAME_DRONE',
        'FRAME_INTERCEPTOR',
        'FRAME_RACER',
        'FRAME_FIGHTER',
        'FRAME_FRIGATE',
        'FRAME_SHUTTLE',
        'FRAME_EXPLORER',
        'FRAME_MINER',
        'FRAME_LIGHT_FREIGHTER',
        'FRAME_HEAVY_FREIGHTER',
        'FRAME_TRANSPORT',
        'FRAME_DESTROYER',
        'FRAME_CRUISER',
        'FRAME_CARRIER',
        'REACTOR_SOLAR_I',
        'REACTOR_FUSION_I',
        'REACTOR_FISSION_I',
        'REACTOR_CHEMICAL_I',
        'REACTOR_ANTIMATTER_I',
        'ENGINE_IMPULSE_DRIVE_I',
        'ENGINE_ION_DRIVE_I',
        'ENGINE_ION_DRIVE_II',
        'ENGINE_HYPER_DRIVE_I',
        'MODULE_MINERAL_PROCESSOR_I',
        'MODULE_GAS_PROCESSOR_I',
        'MODULE_CARGO_HOLD_I',
        'MODULE_CARGO_HOLD_II',
        'MODULE_CARGO_HOLD_III',
        'MODULE_CREW_QUARTERS_I',
        'MODULE_ENVOY_QUARTERS_I',
        'MODULE_PASSENGER_CABIN_I',
        'MODULE_MICRO_REFINERY_I',
        'MODULE_SCIENCE_LAB_I',
        'MODULE_JUMP_DRIVE_I',
        'MODULE_JUMP_DRIVE_II',
        'MODULE_JUMP_DRIVE_III',
        'MODULE_WARP_DRIVE_I',
        'MODULE_WARP_DRIVE_II',
        'MODULE_WARP_DRIVE_III',
        'MODULE_SHIELD_GENERATOR_I',
        'MODULE_SHIELD_GENERATOR_II',
        'MODULE_ORE_REFINERY_I',
        'MODULE_FUEL_REFINERY_I',
        'MOUNT_GAS_SIPHON_I',
        'MOUNT_GAS_SIPHON_II',
        'MOUNT_GAS_SIPHON_III',
        'MOUNT_SURVEYOR_I',
        'MOUNT_SURVEYOR_II',
        'MOUNT_SURVEYOR_III',
        'MOUNT_SENSOR_ARRAY_I',
        'MOUNT_SENSOR_ARRAY_II',
        'MOUNT_SENSOR_ARRAY_III',
        'MOUNT_MINING_LASER_I',
        'MOUNT_MINING_LASER_II',
        'MOUNT_MINING_LASER_III',
        'MOUNT_LASER_CANNON_I',
        'MOUNT_MISSILE_LAUNCHER_I',
        'MOUNT_TURRET_I',
        'SHIP_PROBE',
        'SHIP_MINING_DRONE',
        'SHIP_SIPHON_DRONE',
        'SHIP_INTERCEPTOR',
        'SHIP_LIGHT_HAULER',
        'SHIP_COMMAND_FRIGATE',
        'SHIP_EXPLORER',
        'SHIP_HEAVY_FREIGHTER',
        'SHIP_LIGHT_SHUTTLE',
        'SHIP_ORE_HOUND',
        'SHIP_REFINING_FREIGHTER',
        'SHIP_SURVEYOR',
    ]
)


FLIGHT_MODES = frozenset(['DRIFT', 'STEALTH', 'CRUISE', 'BURN'])


FACTION_SYMBOL = frozenset(
    [
        'COSMIC',
        'VOID',
        'GALACTIC',
        'QUANTUM',
        'DOMINION',
        'ASTRO',
        'CORSAIRS',
        'OBSIDIAN',
        'AEGIS',
        'UNITED',
        'SOLITARY',
        'COBALT',
        'OMEGA',
        'ECHO',
        'LORDS',
        'CULT',
        'ANCIENTS',
        'SHADOW',
        'ETHEREAL',
    ]
)


SURVEYABLE_WAYPOINTS = frozenset(
    ['ASTEROID', 'ASTEROID_FIELD', 'ENGINEERED_ASTEROID']
)


SIPHONABLE_SYMBOLS = frozenset(
    ['LIQUID_HYDROGEN', 'LIQUID_NITROGEN', 'HYDROCARBON']
)

MINABLE_SYMBOLS = frozenset(
    [
        'ALUMINUM_ORE',
        'AMMONIA_ICE',
        'COPPER_ORE',
        'DIAMONDS',  # NOTE: Not mentioned in docs
        'GOLD_ORE',
        'ICE_WATER',
        'IRON_ORE',
        'PLATINUM_ORE',
        'PRECIOUS_STONES',
        'QUARTZ_SAND',
        'SILICON_CRYSTALS',
        'SILVER_ORE',
        'URANITE_ORE',  # NOTE: Not mentioned in docs
    ]
)

REFINABLE_SYMBOLS = frozenset(
    [
        'ALUMINUM',
        'COPPER',
        'FUEL',
        'GOLD',
        'IRON',
        'MERITIUM',
        'PLATINUM',
        'SILVER',
        'URANITE',
    ]
)


SHIP_MOUNTS = frozenset(
    [
        'MOUNT_GAS_SIPHON_I',
        'MOUNT_GAS_SIPHON_II',
        'MOUNT_GAS_SIPHON_III',
        'MOUNT_SURVEYOR_I',
        'MOUNT_SURVEYOR_II',
        'MOUNT_SURVEYOR_III',
        'MOUNT_SENSOR_ARRAY_I',
        'MOUNT_SENSOR_ARRAY_II',
        'MOUNT_SENSOR_ARRAY_III',
        'MOUNT_MINING_LASER_I',
        'MOUNT_MINING_LASER_II',
        'MOUNT_MINING_LASER_III',
        'MOUNT_LASER_CANNON_I',
        'MOUNT_MISSILE_LAUNCHER_I',
        'MOUNT_TURRET_I',
    ]
)
